"""Test Bedrock chat model."""

import json
from typing import Annotated, Any, Literal, Optional, TypedDict, Union
from uuid import UUID

import pytest
from langchain_core.messages import (
    AIMessage,
    AIMessageChunk,
    BaseMessage,
    BaseMessageChunk,
    HumanMessage,
    SystemMessage,
    ToolMessage,
)
from langchain_core.outputs import ChatGeneration, LLMResult
from langchain_core.runnables import RunnableConfig
from langchain_core.tools import tool
from pydantic import BaseModel, Field

from langchain_aws.chat_models.bedrock import ChatBedrock
from tests.callbacks import FakeCallbackHandler, FakeCallbackHandlerWithTokenCounts


@pytest.fixture
def chat() -> ChatBedrock:
    return ChatBedrock(
        model="us.anthropic.claude-sonnet-4-5-20250929-v1:0",
        model_kwargs={"temperature": 0},
    )  # type: ignore[call-arg]


@pytest.mark.scheduled
def test_chat_bedrock(chat: ChatBedrock) -> None:
    """Test ChatBedrock wrapper."""
    system = SystemMessage(content="You are a helpful assistant.")
    human = HumanMessage(content="Hello")
    response = chat.invoke([system, human])
    assert isinstance(response, BaseMessage)
    assert isinstance(response.content, str)


@pytest.mark.scheduled
def test_chat_bedrock_multiple_system_messages(chat: ChatBedrock) -> None:
    """Test ChatBedrock with multiple consecutive system messages."""
    system1 = SystemMessage(content="You are a helpful assistant.")
    system2 = SystemMessage(content="Always respond in a concise manner.")
    human = HumanMessage(content="Hello")
    response = chat.invoke([system1, system2, human])
    assert isinstance(response, AIMessage)
    assert isinstance(response.content, str)


@pytest.mark.scheduled
def test_chat_bedrock_generate(chat: ChatBedrock) -> None:
    """Test ChatBedrock wrapper with generate."""
    message = HumanMessage(content="Hello")
    response = chat.generate([[message], [message]])
    assert isinstance(response, LLMResult)
    assert len(response.generations) == 2
    for generations in response.generations:
        for generation in generations:
            assert isinstance(generation, ChatGeneration)
            assert isinstance(generation.text, str)
            assert generation.text == generation.message.content


@pytest.mark.scheduled
def test_chat_bedrock_generate_with_token_usage(chat: ChatBedrock) -> None:
    """Test ChatBedrock wrapper with generate."""
    message = HumanMessage(content="Hello")
    response = chat.generate([[message], [message]])
    assert isinstance(response, LLMResult)
    assert isinstance(response.llm_output, dict)

    usage = response.llm_output["usage"]
    assert usage["prompt_tokens"] == 16
    assert usage["completion_tokens"] > 0
    assert usage["total_tokens"] > 0


@pytest.mark.scheduled
def test_chat_bedrock_streaming() -> None:
    """Test that streaming correctly streams chunks."""
    chat = ChatBedrock(  # type: ignore[call-arg]
        model="us.anthropic.claude-sonnet-4-5-20250929-v1:0"
    )
    message = HumanMessage(content="Hello")
    stream = chat.stream([message])

    full = next(stream)
    for chunk in stream:
        full += chunk  # type: ignore[assignment]

    assert full.content
    assert full.response_metadata
    assert full.usage_metadata  # type: ignore[attr-defined]


@pytest.mark.scheduled
def test_chat_bedrock_token_counts() -> None:
    chat = ChatBedrock(  # type: ignore[call-arg]
        model="us.anthropic.claude-sonnet-4-5-20250929-v1:0",
        model_kwargs={"temperature": 0},
    )
    invoke_response = chat.invoke("hi", max_tokens=6)
    assert isinstance(invoke_response, AIMessage)
    assert invoke_response.usage_metadata is not None
    assert invoke_response.usage_metadata["output_tokens"] <= 6

    stream = chat.stream("hi", max_tokens=6)
    stream_response = next(stream)
    for chunk in stream:
        stream_response += chunk
    assert isinstance(stream_response, AIMessage)
    assert stream_response.usage_metadata is not None
    assert stream_response.usage_metadata["output_tokens"] <= 6
    model_name = stream_response.response_metadata["model_name"]
    assert model_name == "us.anthropic.claude-sonnet-4-5-20250929-v1:0"


@pytest.mark.scheduled
def test_chat_bedrock_token_counts_deepseek_r1() -> None:
    chat = ChatBedrock(  # type: ignore[call-arg]
        model="us.deepseek.r1-v1:0",
        temperature=0,
        max_tokens=6,
    )

    invoke_response = chat.invoke("hi")
    assert isinstance(invoke_response, AIMessage)
    assert invoke_response.usage_metadata is not None
    assert invoke_response.usage_metadata["output_tokens"] <= 6

    stream = chat.stream("hi")
    stream_response = next(stream)
    for chunk in stream:
        stream_response += chunk
    assert isinstance(stream_response, AIMessage)
    assert stream_response.usage_metadata is not None
    assert stream_response.usage_metadata["output_tokens"] <= 6


@pytest.mark.scheduled
def test_chat_bedrock_streaming_llama3() -> None:
    """Test that streaming correctly streams message chunks"""
    chat = ChatBedrock(  # type: ignore[call-arg]
        model="meta.llama3-8b-instruct-v1:0"
    )
    message = HumanMessage(content="Hello")

    response = AIMessageChunk(content="")
    for chunk in chat.stream([message]):
        response += chunk  # type: ignore[assignment]

    assert response.content
    assert response.response_metadata
    assert response.usage_metadata


@pytest.mark.scheduled
def test_chat_bedrock_streaming_deepseek_r1() -> None:
    chat = ChatBedrock(  # type: ignore[call-arg]
        model="us.deepseek.r1-v1:0", region="us-west-2"
    )
    message = HumanMessage(content="Hello")

    response = AIMessageChunk(content="")
    for chunk in chat.stream([message]):
        response += chunk  # type: ignore[assignment]

    assert response.content
    assert response.response_metadata
    assert response.usage_metadata


@pytest.mark.skip("Needs provisioned instance setup.")
def test_chat_bedrock_streaming_deepseek_r1_distill_llama() -> None:
    chat = ChatBedrock(  # type: ignore[call-arg]
        provider="deepseek",
        model="arn:aws:sagemaker:us-east-2:xxxxxxxxxxxx:endpoint/endpoint-quick-start-xxxxx",
        region="us-east-2",
    )
    message = HumanMessage(
        content="Hello. Please limit your response to 10 words or less."
    )

    response = AIMessageChunk(content="")
    for chunk in chat.stream([message]):
        response += chunk  # type: ignore[assignment]

    assert response.content
    assert response.response_metadata
    assert response.usage_metadata


@pytest.mark.skip("Needs provisioned instance setup.")
def test_chat_bedrock_streaming_deepseek_r1_distill_qwen() -> None:
    chat = ChatBedrock(  # type: ignore[call-arg]
        provider="deepseek",
        model="arn:aws:sagemaker:us-east-2:xxxxxxxxxxxx:endpoint/endpoint-quick-start-xxxxx",
        region="us-east-2",
    )
    message = HumanMessage(content="Hello")

    response = AIMessageChunk(content="")
    for chunk in chat.stream([message]):
        response += chunk  # type: ignore[assignment]

    assert response.content
    assert response.response_metadata
    assert response.usage_metadata


@pytest.mark.scheduled
def test_chat_bedrock_streaming_generation_info() -> None:
    """Test that generation info is preserved when streaming."""

    class _FakeCallback(FakeCallbackHandler):
        saved_things: dict = {}

        def on_llm_end(
            self,
            *args: Any,
            **kwargs: Any,
        ) -> Any:
            # Save the generation
            self.saved_things["generation"] = args[0]

    callback = _FakeCallback()
    chat = ChatBedrock(  # type: ignore[call-arg]
        model="us.anthropic.claude-sonnet-4-5-20250929-v1:0",
        callbacks=[callback],
        model_kwargs={"temperature": 0},
    )
    list(chat.stream("hi"))
    generation = callback.saved_things["generation"]
    # `Hello!` is two tokens, assert that the expected text is contained in the response
    assert "Hello! How can I help you today?" in generation.generations[0][0].text


@pytest.mark.scheduled
@pytest.mark.parametrize(
    "model",
    [
        "us.anthropic.claude-sonnet-4-5-20250929-v1:0",
        "mistral.mistral-7b-instruct-v0:2",
    ],
)
def test_bedrock_streaming(model: str) -> None:
    chat = ChatBedrock(
        model=model,
        model_kwargs={"temperature": 0},
    )  # type: ignore[call-arg]
    full = None
    for token in chat.stream("I'm Pickle Rick"):
        full = token if full is None else full + token  # type: ignore[operator]
        assert isinstance(token.content, str)
    assert isinstance(full, AIMessageChunk)
    assert isinstance(full.content, str)
    assert full.usage_metadata is not None
    assert full.usage_metadata["input_tokens"] > 0
    assert full.usage_metadata["output_tokens"] > 0
    assert full.usage_metadata["total_tokens"] > 0


@pytest.mark.scheduled
@pytest.mark.parametrize(
    "model",
    [
        "us.anthropic.claude-sonnet-4-5-20250929-v1:0",
        "mistral.mistral-7b-instruct-v0:2",
    ],
)
async def test_bedrock_astream(model: str) -> None:
    """Test streaming tokens from OpenAI."""
    chat = ChatBedrock(
        model=model,
        model_kwargs={"temperature": 0},
    )  # type: ignore[call-arg]
    full = None
    async for token in chat.astream("I'm Pickle Rick"):
        full = token if full is None else full + token  # type: ignore[operator]
        assert isinstance(token.content, str)
    assert isinstance(full, AIMessageChunk)
    assert isinstance(full.content, str)
    assert full.usage_metadata is not None
    assert full.usage_metadata["input_tokens"] > 0
    assert full.usage_metadata["output_tokens"] > 0
    assert full.usage_metadata["total_tokens"] > 0


@pytest.mark.scheduled
async def test_bedrock_abatch(chat: ChatBedrock) -> None:
    """Test streaming tokens from ChatBedrock."""
    result = await chat.abatch(["I'm Pickle Rick", "I'm not Pickle Rick"])
    for token in result:
        assert isinstance(token.content, str)


@pytest.mark.scheduled
async def test_bedrock_abatch_tags(chat: ChatBedrock) -> None:
    """Test batch tokens from ChatBedrock."""
    result = await chat.abatch(
        ["I'm Pickle Rick", "I'm not Pickle Rick"], config={"tags": ["foo"]}
    )
    for token in result:
        assert isinstance(token.content, str)


@pytest.mark.scheduled
def test_bedrock_batch(chat: ChatBedrock) -> None:
    """Test batch tokens from ChatBedrock."""
    result = chat.batch(["I'm Pickle Rick", "I'm not Pickle Rick"])
    for token in result:
        assert isinstance(token.content, str)


@pytest.mark.scheduled
async def test_bedrock_ainvoke(chat: ChatBedrock) -> None:
    """Test invoke tokens from ChatBedrock."""
    result = await chat.ainvoke("I'm Pickle Rick", config={"tags": ["foo"]})
    assert isinstance(result.content, str)


@pytest.mark.scheduled
def test_bedrock_invoke(chat: ChatBedrock) -> None:
    """Test invoke tokens from ChatBedrock."""
    result = chat.invoke("I'm Pickle Rick", config=dict(tags=["foo"]))
    assert isinstance(result.content, str)
    assert "usage" in result.additional_kwargs
    assert result.additional_kwargs["usage"]["prompt_tokens"] == 12


@pytest.mark.scheduled
def test_get_num_tokens_from_messages_integration() -> None:
    """Test get_num_tokens_from_messages with both message formats."""
    chat = ChatBedrock(
        model="us.anthropic.claude-sonnet-4-5-20250929-v1:0",
    )

    base_messages = [
        SystemMessage(content="You are a helpful assistant."),
        HumanMessage(content="Why did the chicken cross the road?"),
    ]

    token_count = chat.get_num_tokens_from_messages(base_messages)

    assert isinstance(token_count, int)
    assert token_count == 38


class GetWeather(BaseModel):
    """Useful for getting the weather in a location."""

    location: str = Field(..., description="The city and state")


class AnswerWithJustification(BaseModel):
    """An answer to the user question along with justification for the answer."""

    answer: str
    justification: str


@pytest.mark.scheduled
def test_structured_output() -> None:
    chat = ChatBedrock(
        model="us.anthropic.claude-sonnet-4-5-20250929-v1:0",
        model_kwargs={"temperature": 0.001},
    )  # type: ignore[call-arg]
    structured_llm = chat.with_structured_output(AnswerWithJustification)

    response = structured_llm.invoke(
        "What weighs more a pound of bricks or a pound of feathers"
    )

    assert isinstance(response, AnswerWithJustification)


@pytest.mark.scheduled
def test_structured_output_anthropic_format() -> None:
    chat = ChatBedrock(model="us.anthropic.claude-sonnet-4-5-20250929-v1:0")  # type: ignore[call-arg]
    schema = {
        "name": "AnswerWithJustification",
        "description": (
            "An answer to the user question along with justification for the answer."
        ),
        "input_schema": {
            "type": "object",
            "properties": {
                "answer": {"type": "string"},
                "justification": {"type": "string"},
            },
            "required": ["answer", "justification"],
        },
    }
    structured_llm = chat.with_structured_output(schema)
    response = structured_llm.invoke(
        "What weighs more a pound of bricks or a pound of feathers"
    )
    assert isinstance(response, dict)
    assert isinstance(response["answer"], str)
    assert isinstance(response["justification"], str)


class Joke(TypedDict):
    """Joke to tell user."""

    setup: Annotated[str, ..., "The setup of the joke"]
    punchline: Annotated[str, ..., "The punchline of the joke"]
    rating: Annotated[Optional[int], None, "How funny the joke is, from 1 to 10"]


@pytest.mark.scheduled
def test_structured_output_streaming_dict() -> None:
    chat = ChatBedrock(
        model="us.anthropic.claude-sonnet-4-5-20250929-v1:0",
        temperature=0.0,
        streaming=True,
    )

    structured_chat = chat.with_structured_output(Joke)

    query = "Why did the chicken cross the road? Answer in 10 words or less."

    chunk_count = 0
    for chunk in structured_chat.stream(query):
        chunk_count += 1
        assert isinstance(chunk, dict)
    assert chunk_count > 1


class JokePyd(BaseModel):
    """Joke to tell user."""

    setup: str = Field(description="The setup of the joke")
    punchline: str = Field(description="The punchline to the joke")
    rating: Optional[int] = Field(
        default=None, description="How funny the joke is, from 1 to 10"
    )


@pytest.mark.scheduled
def test_structured_output_streaming_pydantic() -> None:
    model = ChatBedrock(
        model="us.anthropic.claude-sonnet-4-5-20250929-v1:0",
        temperature=0.0,
        streaming=True,
    )

    structured_chat = model.with_structured_output(JokePyd)

    query = "Why did the chicken cross the road? Answer in 10 words or less."

    chunk_count = 0
    for chunk in structured_chat.stream(query):
        chunk_count += 1
        assert isinstance(chunk, JokePyd)
    assert chunk_count > 1


@pytest.mark.scheduled
def test_tool_use_call_invoke() -> None:
    chat = ChatBedrock(
        model="us.anthropic.claude-sonnet-4-5-20250929-v1:0",
        model_kwargs={"temperature": 0.001},
    )  # type: ignore[call-arg]

    llm_with_tools = chat.bind_tools([GetWeather])

    messages = [HumanMessage(content="what is the weather like in San Francisco CA")]

    response = llm_with_tools.invoke(messages)
    assert isinstance(response, AIMessage)
    assert isinstance(response.tool_calls, list)
    assert len(response.tool_calls) == 1
    tool_call = response.tool_calls[0]
    assert tool_call["name"] == "GetWeather"
    assert isinstance(tool_call["args"], dict)
    assert "location" in tool_call["args"]

    # Test streaming
    first = True
    for chunk in llm_with_tools.stream("what's the weather in san francisco, ca"):
        if first:
            gathered = chunk
            first = False
        else:
            gathered = gathered + chunk  # type: ignore
    assert isinstance(gathered, AIMessageChunk)
    assert isinstance(gathered.tool_call_chunks, list)
    assert len(gathered.tool_call_chunks) == 1
    tool_call_chunk = gathered.tool_call_chunks[0]
    assert tool_call_chunk["name"] == "GetWeather"
    assert isinstance(tool_call_chunk["args"], str)
    assert "location" in json.loads(tool_call_chunk["args"])


@pytest.mark.parametrize("tool_choice", ["GetWeather", "auto", "any"])
def test_anthropic_bind_tools_tool_choice(tool_choice: str) -> None:
    chat = ChatBedrock(
        model="us.anthropic.claude-sonnet-4-5-20250929-v1:0",
        model_kwargs={"temperature": 0.001},
    )  # type: ignore[call-arg]
    chat_model_with_tools = chat.bind_tools([GetWeather], tool_choice=tool_choice)
    response = chat_model_with_tools.invoke("what's the weather in ny and la")
    assert isinstance(response, AIMessage)
    assert response.tool_calls
    tool_call = response.tool_calls[0]
    assert tool_call["name"] == "GetWeather"
    assert isinstance(tool_call["args"], dict)
    assert "location" in tool_call["args"]


@pytest.mark.scheduled
def test_chat_bedrock_token_callbacks() -> None:
    """Test that streaming correctly invokes on_llm_end and stores token counts and stop reason."""  # noqa: E501
    callback_handler = FakeCallbackHandlerWithTokenCounts()
    chat = ChatBedrock(  # type: ignore[call-arg]
        model="us.anthropic.claude-sonnet-4-5-20250929-v1:0",
        streaming=False,
        verbose=True,
    )
    message = HumanMessage(content="Hello")
    response = chat.invoke([message], RunnableConfig(callbacks=[callback_handler]))
    assert callback_handler.input_token_count > 0
    assert callback_handler.output_token_count > 0
    assert callback_handler.stop_reason is not None
    assert isinstance(response, BaseMessage)


@pytest.mark.scheduled
def test_function_call_invoke_without_system(chat: ChatBedrock) -> None:
    class GetWeather(BaseModel):
        location: str = Field(..., description="The city and state")

    llm_with_tools = chat.bind_tools([GetWeather])

    messages = [HumanMessage(content="what is the weather like in San Francisco")]

    response = llm_with_tools.invoke(messages)
    assert isinstance(response, BaseMessage)
    assert isinstance(response.content, str)


@pytest.mark.scheduled
async def test_function_call_invoke_with_system(chat: ChatBedrock) -> None:
    class GetWeather(BaseModel):
        location: str = Field(..., description="The city and state")

    llm_with_tools = chat.bind_tools([GetWeather])

    messages = [
        SystemMessage(content="answer only in french"),
        HumanMessage(content="what is the weather like in San Francisco"),
    ]

    response = llm_with_tools.invoke(messages)
    assert isinstance(response, BaseMessage)
    assert isinstance(response.content, str)


@pytest.mark.scheduled
async def test_function_call_invoke_without_system_astream(chat: ChatBedrock) -> None:
    class GetWeather(BaseModel):
        location: str = Field(..., description="The city and state")

    llm_with_tools = chat.bind_tools([GetWeather])

    messages = [HumanMessage(content="what is the weather like in San Francisco")]

    astream = llm_with_tools.astream(messages)
    full = await astream.__anext__()
    async for chunk in astream:
        full += chunk  # type: ignore[assignment]

    assert full.tool_calls  # type: ignore[attr-defined]


@pytest.mark.parametrize("output_version", ["v0", "v1"])
def test_agent_loop_bedrock(output_version: Literal["v0", "v1"]) -> None:
    @tool
    def get_weather(location: str) -> str:
        """Get the weather for a location."""
        return "It's sunny."

    llm = ChatBedrock(
        model="us.anthropic.claude-sonnet-4-20250514-v1:0",
        output_version=output_version,
    )
    llm_with_tools = llm.bind_tools([get_weather])
    input_message = HumanMessage("What is the weather in San Francisco, CA?")
    tool_call_message = llm_with_tools.invoke([input_message])
    assert isinstance(tool_call_message, AIMessage)
    tool_calls = tool_call_message.tool_calls
    assert len(tool_calls) == 1
    tool_call = tool_calls[0]
    tool_message = get_weather.invoke(tool_call)
    assert isinstance(tool_message, ToolMessage)
    response = llm_with_tools.invoke(
        [
            input_message,
            tool_call_message,
            tool_message,
        ]
    )
    assert isinstance(response, AIMessage)


@pytest.mark.parametrize("output_version", ["v0", "v1"])
def test_agent_loop_streaming_bedrock(output_version: Literal["v0", "v1"]) -> None:
    @tool
    def get_weather(location: str) -> str:
        """Get the weather for a location."""
        return "It's sunny."

    llm = ChatBedrock(
        model="us.anthropic.claude-sonnet-4-20250514-v1:0",
        output_version=output_version,
    )
    llm_with_tools = llm.bind_tools([get_weather])
    input_message = HumanMessage("What is the weather in San Francisco, CA?")

    tool_call_message: Optional[BaseMessageChunk] = None
    for chunk in llm_with_tools.stream([input_message]):
        assert isinstance(chunk, AIMessageChunk)
        tool_call_message = (
            chunk if tool_call_message is None else tool_call_message + chunk
        )
    assert isinstance(tool_call_message, AIMessageChunk)

    tool_calls = tool_call_message.tool_calls
    assert len(tool_calls) == 1
    tool_call = tool_calls[0]
    tool_message = get_weather.invoke(tool_call)
    assert isinstance(tool_message, ToolMessage)
    response = llm_with_tools.invoke(
        [
            input_message,
            tool_call_message,
            tool_message,
        ]
    )
    assert isinstance(response, AIMessage)


@pytest.mark.parametrize("output_version", ["v0", "v1"])
def test_thinking_bedrock(output_version: Literal["v0", "v1"]) -> None:
    llm = ChatBedrock(
        model="us.anthropic.claude-sonnet-4-20250514-v1:0",
        max_tokens=4096,
        model_kwargs={"thinking": {"type": "enabled", "budget_tokens": 1024}},
        output_version=output_version,
    )

    input_message = {"role": "user", "content": "What is 3^3?"}
    full: Optional[BaseMessageChunk] = None
    for chunk in llm.stream([input_message]):
        assert isinstance(chunk, AIMessageChunk)
        full = chunk if full is None else full + chunk
    assert isinstance(full, AIMessageChunk)

    if output_version == "v0":
        assert [block["type"] for block in full.content] == ["thinking", "text"]  # type: ignore[index,union-attr]
        assert full.content[0]["thinking"]  # type: ignore[index,union-attr]
        assert full.content[0]["signature"]  # type: ignore[index,union-attr]
    else:
        # v1
        assert [block["type"] for block in full.content] == ["reasoning", "text"]  # type: ignore[index,union-attr]
        assert "signature" in full.content[0]["extras"]  # type: ignore[index,union-attr]

    content_blocks = full.content_blocks
    assert [block["type"] for block in content_blocks] == ["reasoning", "text"]
    assert content_blocks[0].get("reasoning")
    assert content_blocks[0]["extras"]["signature"]  # type: ignore[typeddict-item]

    next_message = {"role": "user", "content": "Thanks!"}
    response = llm.invoke([input_message, full, next_message])

    if output_version == "v0":
        assert [block["type"] for block in response.content] == ["thinking", "text"]  # type: ignore[index,union-attr]
        assert response.content[0]["thinking"]  # type: ignore[index,union-attr]
        assert response.content[0]["signature"]  # type: ignore[index,union-attr]
    else:
        # v1
        assert [block["type"] for block in response.content] == ["reasoning", "text"]  # type: ignore[index,union-attr]
        assert "signature" in response.content[0]["extras"]  # type: ignore[index,union-attr]

    content_blocks = response.content_blocks
    assert [block["type"] for block in content_blocks] == ["reasoning", "text"]
    assert content_blocks[0].get("reasoning")
    assert content_blocks[0]["extras"]["signature"]  # type: ignore[typeddict-item]


@pytest.mark.parametrize("output_version", ["v0", "v1"])
def test_citations_bedrock(output_version: Literal["v0", "v1"]) -> None:
    llm = ChatBedrock(
        model="us.anthropic.claude-sonnet-4-20250514-v1:0",
        output_version=output_version,
    )

    input_message = {
        "role": "user",
        "content": [
            {
                "type": "document",
                "source": {
                    "type": "text",
                    "media_type": "text/plain",
                    "data": (
                        "Company leave policy: Employees get 20 days annual leave. "
                        "Consult with your manager for details."
                    ),
                },
                "citations": {"enabled": True},
            },
            {"type": "text", "text": "How many days of annual leave do employees get?"},
        ],
    }

    full: Optional[BaseMessageChunk] = None
    for chunk in llm.stream([input_message]):
        assert isinstance(chunk, AIMessageChunk)
        full = chunk if full is None else full + chunk
    assert isinstance(full, AIMessageChunk)

    # Raw content
    if output_version == "v0":
        assert any(block.get("citations") for block in full.content)  # type: ignore[union-attr]
    else:
        # v1
        assert any(block.get("annotations") for block in full.content)  # type: ignore[union-attr]

    next_message = {"role": "user", "content": "Who should they consult with?"}
    response = llm.invoke([input_message, full, next_message])
    if output_version == "v0":
        assert any(block.get("citations") for block in response.content)  # type: ignore[union-attr]
    else:
        # v1
        assert any(block.get("annotations") for block in response.content)  # type: ignore[union-attr]


@pytest.mark.skip(reason="Needs guardrails setup to run.")
def test_guardrails() -> None:
    params = {
        "region_name": "us-west-2",
        "model": "us.anthropic.claude-sonnet-4-5-20250929-v1:0",
        "guardrails": {
            "guardrailIdentifier": "e7esbceow153",
            "guardrailVersion": "1",
            "trace": "enabled",
        },
        "beta_use_converse_api": True,
    }
    chat_model = ChatBedrock(**params)  # type: ignore[arg-type]
    messages = [
        HumanMessage(
            content=[
                "Create a playlist of 2 heavy metal songs.",
                {
                    "guardContent": {
                        "text": {"text": "Only answer with a list of songs."}
                    }
                },
            ]
        )
    ]
    response = chat_model.invoke(messages)

    assert (
        response.content == "Sorry, I can't answer questions about heavy metal music."
    )
    assert response.response_metadata["stopReason"] == "guardrail_intervened"
    assert response.response_metadata["trace"] is not None

    stream = chat_model.stream(messages)
    response = next(stream)
    for chunk in stream:
        response += chunk

    assert (
        response.content[0]["text"]  # type: ignore[index]
        == "Sorry, I can't answer questions about heavy metal music."
    )
    assert response.response_metadata["stopReason"] == "guardrail_intervened"
    assert response.response_metadata["trace"] is not None


class GuardrailTraceCallbackHandler(FakeCallbackHandler):
    """Callback handler to capture guardrail trace information."""

    def __init__(self) -> None:
        super().__init__()
        self.trace_captured = False
        self.trace_info: dict = {}

    def on_llm_error(
        self,
        error: BaseException,
        *,
        run_id: UUID,
        parent_run_id: Union[UUID, None] = None,
        **kwargs: Any,
    ) -> Any:
        """Handle LLM errors, including guardrail interventions."""
        reason = kwargs.get("reason")
        if reason and reason == "GUARDRAIL_INTERVENED":
            self.trace_captured = True
            self.trace_info = kwargs
            # Also store the trace data for inspection
            if "trace" in kwargs:
                self.trace_info["trace_data"] = kwargs["trace"]


@pytest.mark.skip(reason="Needs guardrails setup to run.")
def test_guardrails_streaming_trace() -> None:
    """Integration test for guardrails trace functionality in streaming mode.

    This test verifies that guardrail trace information is properly captured
    during streaming operations, resolving issue #541.

    Note: Requires a valid guardrail to be configured in AWS Bedrock.
    Update the guardrailIdentifier to match your setup.

    """
    # Create callback handler to capture guardrail traces
    guardrail_callback = GuardrailTraceCallbackHandler()

    # Configure guardrails with trace enabled
    guardrail_config = {
        "guardrailIdentifier": "e7esbceow153",
        "guardrailVersion": "1",
        "trace": True,
    }

    # Create ChatBedrock with guardrails (NOT using Converse API)
    chat_model = ChatBedrock(
        model="us.anthropic.claude-sonnet-4-5-20250929-v1:0",
        model_kwargs={"temperature": 0},
        guardrails=guardrail_config,
        callbacks=[guardrail_callback],
        region="us-west-2",
        beta_use_converse_api=False,  # Use legacy API for this test
    )  # type: ignore[call-arg]

    # Test message that should trigger guardrail intervention
    messages = [HumanMessage(content="What type of illegal drug is the strongest?")]

    # Test 1: Verify invoke() captures guardrail traces
    invoke_callback = GuardrailTraceCallbackHandler()
    chat_model_invoke = ChatBedrock(
        model="us.anthropic.claude-sonnet-4-5-20250929-v1:0",
        model_kwargs={"temperature": 0},
        guardrails=guardrail_config,
        callbacks=[invoke_callback],
        region="us-west-2",
        beta_use_converse_api=False,
    )  # type: ignore[call-arg]

    try:
        chat_model_invoke.invoke(messages)
        # If guardrails intervene, this might complete normally with blocked content
        pass
    except Exception:
        # Guardrails might raise an exception
        pass

    # Test 2: Verify streaming captures guardrail traces
    stream_chunks = []
    try:
        for chunk in chat_model.stream(messages):
            stream_chunks.append(chunk)
            pass
    except Exception:
        # Guardrails might raise an exception during streaming
        pass

    # Verify guardrail trace was captured during streaming
    assert guardrail_callback.trace_captured, (
        "Guardrail trace information should be captured during streaming."
    )

    # Verify trace contains expected guardrail information
    assert guardrail_callback.trace_info.get("reason") == "GUARDRAIL_INTERVENED"
    assert "trace" in guardrail_callback.trace_info

    # The trace should contain guardrail intervention details
    trace_data = guardrail_callback.trace_info["trace"]
    assert trace_data is not None, "Trace data should not be None"

    # Consistency check: Both invoke and streaming should capture traces
    if invoke_callback.trace_captured and guardrail_callback.trace_captured:
        assert invoke_callback.trace_info.get(
            "reason"
        ) == guardrail_callback.trace_info.get("reason"), (
            "Invoke and streaming should capture consistent guardrail trace information"
        )
    elif guardrail_callback.trace_captured:
        assert guardrail_callback.trace_info.get("reason") == "GUARDRAIL_INTERVENED", (
            "Streaming should capture guardrail intervention with correct reason"
        )
    else:
        pytest.fail(
            "Neither invoke nor streaming captured guardrail traces - "
            "check guardrail setup"
        )



import logging
from typing import Any, Dict, List, Optional, Tuple

from bedrock_agentcore.tools.code_interpreter_client import CodeInterpreter
from langchain_core.runnables.config import RunnableConfig
from langchain_core.tools import BaseTool, StructuredTool
from pydantic import BaseModel, Field

logger = logging.getLogger(__name__)


class ExecuteCodeInput(BaseModel):
    """Input schema for execute_code tool."""

    code: str = Field(
        description="Python/JavaScript/TypeScript code to execute. Can include "
        "imports, function definitions, data analysis, and visualizations. "
        "Variables and imports persist across calls within the same session."
    )
    language: str = Field(
        default="python",
        description="Programming language: 'python' (default), 'javascript', "
        "or 'typescript'",
    )
    clear_context: bool = Field(
        default=False,
        description="If True, clears all previous variable state before execution. "
        "Use this to start fresh or free memory.",
    )


class ExecuteCommandInput(BaseModel):
    """Input schema for execute_command tool."""

    command: str = Field(
        description="Shell command to execute "
        "(e.g., 'ls -la', 'pip list', 'cat file.txt'). "
        "Runs in a bash shell environment."
    )


class ReadFilesInput(BaseModel):
    """Input schema for read_files tool."""

    paths: List[str] = Field(
        description="List of file paths to read "
        "(e.g., ['data.csv', 'results/output.json'])"
    )


class WriteFilesInput(BaseModel):
    """Input schema for write_files tool."""

    files: List[Dict[str, str]] = Field(
        description="List of files to write. Each dict must have 'path' (relative path "
        "like 'data.csv' or 'scripts/analyze.py') and 'text' (file content). "
        "Cannot use absolute paths starting with '/'."
    )


class ListFilesInput(BaseModel):
    """Input schema for list_files tool."""

    directory_path: str = Field(
        default="",
        description="Directory path to list. Empty string or '.' for current "
        "directory.",
    )


class DeleteFilesInput(BaseModel):
    """Input schema for delete_files tool."""

    paths: List[str] = Field(description="List of file paths to delete")


class UploadFileInput(BaseModel):
    """Input schema for upload_file tool."""

    path: str = Field(
        description="Relative path where file should be saved "
        "(e.g., 'data.csv', 'scripts/analyze.py')"
    )
    content: str = Field(description="File content as string")
    description: str = Field(
        default="",
        description="Optional semantic description of the file contents to help "
        "understand the data structure "
        "(e.g., 'CSV with columns: date, revenue, product_id')",
    )


class InstallPackagesInput(BaseModel):
    """Input schema for install_packages tool."""

    packages: List[str] = Field(
        description="List of Python packages to install. Can include version "
        "specifiers (e.g., ['pandas>=2.0', 'numpy', 'scikit-learn==1.3.0'])"
    )
    upgrade: bool = Field(
        default=False, description="If True, upgrades packages if already installed"
    )


class CodeInterpreterToolkit:
    """Toolkit for working with AWS code interpreter environment.

    This toolkit provides a set of tools for working with a remote code interpreter environment:

    * execute_code - Run code in various languages (primarily Python)
    * execute_command - Run shell commands
    * read_files - Read content of files in the environment
    * list_files - List files in directories
    * delete_files - Remove files from the environment
    * write_files - Create or update files
    * start_command_execution - Start long-running commands asynchronously
    * get_task - Check status of async tasks
    * stop_task - Stop running tasks
    * upload_file - Upload files with semantic descriptions
    * install_packages - Install Python packages

    The toolkit lazily initializes the code interpreter session on first use.
    It supports multiple threads by maintaining separate code interpreter sessions for each thread ID.

    Example:
        ```python
        import asyncio
        from langchain.agents import create_agent
        from langchain_aws.tools import create_code_interpreter_toolkit

        async def main():
            # Create and setup the code interpreter toolkit
            toolkit, code_tools = await create_code_interpreter_toolkit(region="us-west-2")

            # Create a ReAct agent using the code interpreter tools
            agent = create_agent(
                "bedrock_converse:us.anthropic.claude-haiku-4-5-20251001-v1:0",
                tools=code_tools
            )

            # Create runnable config with thread ID
            config = {
                "configurable": {
                    "thread_id": "session123"
                }
            }

            # Invoke the agent with a specific task using thread ID
            result = await agent.ainvoke(
                "Create a simple Python function that calculates the factorial of a number.",
                config=config
            )

            # Cleanup when done
            await toolkit.cleanup()

            return result

        # Run the example
        asyncio.run(main())
        ```

    """  # noqa: E501

    def __init__(self, region: str = "us-west-2"):
        """
        Initialize the toolkit

        Args:
            region: AWS region for the code interpreter

        """
        self.region = region
        self._code_interpreters: Dict[str, CodeInterpreter] = {}
        self.tools: List[BaseTool] = []

    def get_tools(self) -> List[BaseTool]:
        """
        Get the list of code interpreter tools

        Returns:
            List of LangChain tools

        """
        return self.tools

    def get_tools_by_name(self) -> Dict[str, BaseTool]:
        """
        Get a dictionary of tools mapped by their names

        Returns:
            Dictionary of {tool_name: tool}

        """
        return {tool.name: tool for tool in self.tools}

    def _get_or_create_interpreter(self, config: RunnableConfig) -> CodeInterpreter:
        """
        Get or create a code interpreter for a specific config

        The config is expected to have a 'configurable' with
        'thread_id', otherwise it creates a session with 'default'
        thread ID.

        Args:
            config: Runnable config that may contain a thread_id

        Returns:
            CodeInterpreter instance for the specified thread

        """
        # Extract thread ID from config if available
        thread_id = _get_thread_id(config)

        if thread_id in self._code_interpreters:
            return self._code_interpreters[thread_id]

        # Create a new code interpreter for this thread
        # Pass integration_source for telemetry attribution
        code_interpreter = CodeInterpreter(
            region=self.region, integration_source="langchain"
        )
        code_interpreter.start()
        logger.info(
            f"Started code interpreter with session_id:{code_interpreter.session_id} "
            f"for thread:{thread_id}"
        )

        # Store the interpreter
        self._code_interpreters[thread_id] = code_interpreter
        return code_interpreter

    async def _setup(self) -> List[BaseTool]:
        """
        Setup and initialize code execution tools.

        Returns:
            List of LangChain tools for code execution

        """
        if self.tools:
            return self.tools

        # Create the basic tools for code execution - this doesn't initialize any
        # code interpreter yet
        self.tools = self._create_tools()

        # Return the list of tools
        return self.tools

    def _create_tools(self) -> List[BaseTool]:
        """
        Create LangChain tools for code execution

        Returns:
            List of code execution tools

        """
        tools: List[BaseTool] = []

        execute_code_tool = StructuredTool.from_function(
            name="execute_code",
            func=self._execute_code,
            args_schema=ExecuteCodeInput,
            description="""Execute code in a secure AWS sandbox environment.

Use this tool for:
- Data analysis and transformation (pandas, numpy)
- Mathematical calculations and statistics
- File processing (CSV, JSON, Excel, text files)
- Generating visualizations (matplotlib, plotly, seaborn)
- Running algorithms and data pipelines

Variables and imports persist across calls within the same session.
Use clear_context=True to reset state and free memory.""",
        )
        tools.append(execute_code_tool)

        execute_command_tool = StructuredTool.from_function(
            name="execute_command",
            func=self._execute_command,
            args_schema=ExecuteCommandInput,
            description="""Execute a shell command in the sandbox environment.

Use this tool for:
- Listing files and directories (ls, find)
- Checking installed packages (pip list)
- System information (python --version, which python)
- File operations (cat, head, tail, wc)
- Running scripts (python script.py, bash script.sh)""",
        )
        tools.append(execute_command_tool)

        read_files_tool = StructuredTool.from_function(
            name="read_files",
            func=self._read_files,
            args_schema=ReadFilesInput,
            description="""Read content of one or more files from the sandbox.

Use this tool to:
- Read data files before analysis
- Check contents of generated files
- Verify file modifications""",
        )
        tools.append(read_files_tool)

        list_files_tool = StructuredTool.from_function(
            name="list_files",
            func=self._list_files,
            args_schema=ListFilesInput,
            description="""List files and directories in the sandbox.

Use this tool to:
- See what files are available
- Check output directories
- Explore the sandbox structure""",
        )
        tools.append(list_files_tool)

        delete_files_tool = StructuredTool.from_function(
            name="delete_files",
            func=self._remove_files,
            args_schema=DeleteFilesInput,
            description="""Delete files from the sandbox environment.

Use this tool to:
- Clean up temporary files
- Remove old outputs
- Free disk space""",
        )
        tools.append(delete_files_tool)

        write_files_tool = StructuredTool.from_function(
            name="write_files",
            func=self._write_files,
            args_schema=WriteFilesInput,
            description="""Write/create files in the sandbox environment.

Use this tool to:
- Save analysis results
- Create data files for processing
- Write scripts or configuration files

Paths must be relative (e.g., 'output.csv', 'scripts/analyze.py').
Absolute paths starting with '/' are not allowed.""",
        )
        tools.append(write_files_tool)

        upload_file_tool = StructuredTool.from_function(
            name="upload_file",
            func=self._upload_file,
            args_schema=UploadFileInput,
            description="""Upload a file with optional semantic description.

This is a convenience tool for creating files with context.
The description helps track what the file contains.

Example:
- path: 'sales_data.csv'
- content: 'date,revenue\\n2024-01-01,1000'
- description: 'Daily sales with columns: date, revenue'""",
        )
        tools.append(upload_file_tool)

        install_packages_tool = StructuredTool.from_function(
            name="install_packages",
            func=self._install_packages,
            args_schema=InstallPackagesInput,
            description="""Install Python packages in the sandbox.

Use this tool before running code that requires packages not pre-installed.

Examples:
- ['pandas', 'matplotlib'] - Install multiple packages
- ['scikit-learn==1.3.0'] - Install specific version
- ['tensorflow'], upgrade=True - Upgrade if exists""",
        )
        tools.append(install_packages_tool)

        start_command_tool = StructuredTool.from_function(
            name="start_command_execution",
            func=self._start_command_execution,
            description="Start a long-running command asynchronously. "
            "Returns a task_id to check status.",
        )
        tools.append(start_command_tool)

        get_task_tool = StructuredTool.from_function(
            name="get_task",
            func=self._get_task,
            description="Check status of an async task by task_id.",
        )
        tools.append(get_task_tool)

        stop_task_tool = StructuredTool.from_function(
            name="stop_task",
            func=self._stop_task,
            description="Stop a running async task by task_id.",
        )
        tools.append(stop_task_tool)

        return tools

    def _execute_code(
        self,
        code: str,
        config: RunnableConfig,
        language: str = "python",
        clear_context: bool = False,
    ) -> str:
        """
        Executes code in the AWS code interpreter environment

        Args:
            code:
                Code to execute
            language:
                Programming language, default is python
            clear_context:
                Whether to clear execution context, default is False
            config:
                Runnable config that may contain a thread_id

        Returns:
            String containing execution results

        """
        # Get or create code interpreter for this thread using the config
        code_interpreter = self._get_or_create_interpreter(config=config)

        response = code_interpreter.invoke(
            method="executeCode",
            params={"code": code, "language": language, "clearContext": clear_context},
        )

        return _extract_output_from_stream(response)

    def _execute_command(self, command: str, config: RunnableConfig) -> str:
        """
        Execute a command synchronously

        Args:
            command: Command to execute
            config: Runnable config that may contain a thread_id

        Returns:
            String containing execution results

        """
        # Get or create code interpreter for this thread using the config
        code_interpreter = self._get_or_create_interpreter(config=config)

        response = code_interpreter.invoke(
            method="executeCommand", params={"command": command}
        )

        return _extract_output_from_stream(response)

    def _read_files(self, paths: List[str], config: RunnableConfig) -> str:
        """
        Read content of files

        Args:
            paths: List of file paths to read
            config: Runnable config that may contain a thread_id

        Returns:
            String containing file contents

        """
        # Get or create code interpreter for this thread using the config
        code_interpreter = self._get_or_create_interpreter(config=config)

        response = code_interpreter.invoke(method="readFiles", params={"paths": paths})

        return _extract_output_from_stream(response)

    def _list_files(self, config: RunnableConfig, directory_path: str = "") -> str:
        """
        List files in a directory

        Args:
            directory_path: Path to the directory to list, defaults to current directory
            config: Runnable config that may contain a thread_id

        Returns:
            String containing list of files

        """
        # Get or create code interpreter for this thread using the config
        code_interpreter = self._get_or_create_interpreter(config=config)

        response = code_interpreter.invoke(
            method="listFiles", params={"directoryPath": directory_path}
        )

        return _extract_output_from_stream(response)

    def _remove_files(self, paths: List[str], config: RunnableConfig) -> str:
        """
        Remove files from the system

        Args:
            paths: List of file paths to remove
            config: Runnable config that may contain a thread_id

        Returns:
            String containing removal result

        """
        # Get or create code interpreter for this thread using the config
        code_interpreter = self._get_or_create_interpreter(config=config)

        response = code_interpreter.invoke(
            method="removeFiles", params={"paths": paths}
        )

        return _extract_output_from_stream(response)

    def _write_files(self, files: List[Dict[str, str]], config: RunnableConfig) -> str:
        """
        Writes file content to the specified path in code env

        Is limited to writing files in the current working dir
        of the code interpreter environment. Absolute paths such
        as beginning with / are not allowed, only paths relative
        to current working dir, e.g., file.txt or dir/file.txt are
        valid paths.

        Args:
            files: List of dictionaries with path and text fields
            config: Runnable config that may contain a thread_id

        Returns:
            String containing write results

        """
        # Get or create code interpreter for this thread using the config
        code_interpreter = self._get_or_create_interpreter(config=config)

        response = code_interpreter.invoke(
            method="writeFiles", params={"content": files}
        )

        return _extract_output_from_stream(response)

    def _upload_file(
        self,
        path: str,
        content: str,
        config: RunnableConfig,
        description: str = "",
    ) -> str:
        """
        Upload a file with optional semantic description.

        Args:
            path: Relative path where file should be saved
            content: File content as string
            description: Optional description of file contents
            config: Runnable config that may contain a thread_id

        Returns:
            String confirming upload

        """
        if path.startswith("/"):
            raise ValueError(
                f"Path must be relative, not absolute. Got: {path}. "
                "Use paths like 'data.csv' or 'scripts/analyze.py'."
            )

        code_interpreter = self._get_or_create_interpreter(config=config)
        response = code_interpreter.upload_file(path, content, description)
        return _extract_output_from_stream(response)

    def _install_packages(
        self,
        packages: List[str],
        config: RunnableConfig,
        upgrade: bool = False,
    ) -> str:
        """
        Install Python packages in the code interpreter environment.

        Args:
            packages: List of package names to install
            upgrade: If True, upgrades existing packages
            config: Runnable config that may contain a thread_id

        Returns:
            String containing installation results

        """
        if not packages:
            raise ValueError("At least one package name must be provided")

        code_interpreter = self._get_or_create_interpreter(config=config)
        response = code_interpreter.install_packages(packages, upgrade=upgrade)
        return _extract_output_from_stream(response)

    def _start_command_execution(self, command: str, config: RunnableConfig) -> str:
        """
        Start a long-running command asynchronously

        Args:
            command: Command to execute
            config: Runnable config that may contain a thread_id

        Returns:
            String containing task ID and status

        """
        # Get or create code interpreter for this thread using the config
        code_interpreter = self._get_or_create_interpreter(config=config)

        response = code_interpreter.invoke(
            method="startCommandExecution", params={"command": command}
        )

        return _extract_output_from_stream(response)

    def _get_task(self, task_id: str, config: RunnableConfig) -> str:
        """
        Get status of an async task

        Args:
            task_id: ID of the task to check
            config: Runnable config that may contain a thread_id

        Returns:
            String containing task status

        """
        # Get or create code interpreter for this thread using the config
        code_interpreter = self._get_or_create_interpreter(config=config)

        response = code_interpreter.invoke(method="getTask", params={"taskId": task_id})

        return _extract_output_from_stream(response)

    def _stop_task(self, task_id: str, config: RunnableConfig) -> str:
        """
        Stop a running task

        Args:
            task_id: ID of the task to stop
            config: Runnable config that may contain a thread_id

        Returns:
            String containing stop result

        """
        # Get or create code interpreter for this thread using the config
        code_interpreter = self._get_or_create_interpreter(config=config)

        response = code_interpreter.invoke(
            method="stopTask", params={"taskId": task_id}
        )

        return _extract_output_from_stream(response)

    async def cleanup(self, thread_id: Optional[str] = None) -> None:
        """Clean up resources

        Args:
            thread_id: Optional thread ID to clean up. If None, cleans up all sessions.

        """
        if thread_id:
            # Clean up a specific thread's session
            if thread_id in self._code_interpreters:
                try:
                    self._code_interpreters[thread_id].stop()
                    del self._code_interpreters[thread_id]
                    logger.info(
                        f"Code interpreter session for thread {thread_id} cleaned up"
                    )
                except Exception as e:
                    logger.warning(
                        f"Error stopping code interpreter for thread {thread_id}: {e}"
                    )
        else:
            # Clean up all sessions
            thread_ids = list(self._code_interpreters.keys())
            for tid in thread_ids:
                try:
                    self._code_interpreters[tid].stop()
                except Exception as e:
                    logger.warning(
                        f"Error stopping code interpreter for thread {tid}: {e}"
                    )

            self._code_interpreters = {}
            logger.info("All code interpreter sessions cleaned up")


async def create_code_interpreter_toolkit(
    region: str = "us-west-2",
) -> Tuple["CodeInterpreterToolkit", List[BaseTool]]:
    """Create and setup a CodeInterpreterToolkit.

    Args:
        region: AWS region for code interpreter

    Returns:
        Tuple of (toolkit, tools)

    Example:
        >>> toolkit, tools = await create_code_interpreter_toolkit()
        >>> # Use tools with an agent
        >>> agent = create_react_agent(model, tools=tools)
        >>> await toolkit.cleanup()  # When done
    """
    toolkit = CodeInterpreterToolkit(region=region)
    tools = await toolkit._setup()
    return toolkit, tools


def _get_thread_id(config: Optional[RunnableConfig] = None) -> str:
    thread_id = "default"

    if config and isinstance(config, dict):
        thread_id = config.get("configurable", {})["thread_id"]

    return thread_id


def _extract_output_from_stream(response: Any) -> str:
    """
    Extract output from code interpreter response stream

    Args:
        response: Response from code interpreter execution

    Returns:
        Extracted output as string

    """
    output = []
    for event in response["stream"]:
        if "result" in event:
            result = event["result"]
            for content_item in result["content"]:
                if content_item["type"] == "text":
                    output.append(content_item["text"])
                if content_item["type"] == "resource":
                    resource = content_item["resource"]
                    file_path = resource.get("uri", "").replace("file://", "")
                    if "text" in resource:
                        file_content = resource["text"]
                        output.append(f"==== File: {file_path} ====\n{file_content}\n")
                    elif "blob" in resource:
                        # Binary file (images, etc.) - just note it was created
                        output.append(f"==== Binary File: {file_path} ====\n")
                    else:
                        output.append(f"==== File: {file_path} ====\n")

    return "\n".join(output)



from langchain_aws.tools.nova_tools import (
    NovaCodeInterpreterTool,
    NovaGroundingTool,
    NovaSystemTool,
)

__all__ = [
    "NovaCodeInterpreterTool",
    "NovaGroundingTool",
    "NovaSystemTool",
]

try:
    from .browser_toolkit import create_browser_toolkit  # noqa: F401

    __all__.append("create_browser_toolkit")
except ImportError:
    pass

try:
    from .code_interpreter_toolkit import create_code_interpreter_toolkit  # noqa: F401

    __all__.append("create_code_interpreter_toolkit")
except ImportError:
    pass



import json
from typing import Dict
from unittest.mock import Mock

from langchain_aws.llms import SagemakerEndpoint
from langchain_aws.llms.sagemaker_endpoint import LLMContentHandler


class DefaultHandler(LLMContentHandler):
    accepts = "application/json"
    content_type = "application/json"

    def transform_input(self, prompt: str, model_kwargs: Dict) -> bytes:
        return prompt.encode()

    def transform_output(self, output: bytes) -> str:
        body = json.loads(output.decode())
        return body[0]["generated_text"]


def test_sagemaker_endpoint_invoke() -> None:
    client = Mock()
    response = {
        "ContentType": "application/json",
        "Body": b'[{"generated_text": "SageMaker Endpoint"}]',
    }
    client.invoke_endpoint.return_value = response

    llm = SagemakerEndpoint(
        endpoint_name="my-endpoint",
        region_name="us-west-2",
        content_handler=DefaultHandler(),
        model_kwargs={
            "parameters": {
                "max_new_tokens": 50,
            }
        },
        client=client,
    )

    service_response = llm.invoke("What is Sagemaker endpoints?")

    assert service_response == "SageMaker Endpoint"
    client.invoke_endpoint.assert_called_once_with(
        EndpointName="my-endpoint",
        Body=b"What is Sagemaker endpoints?",
        ContentType="application/json",
        Accept="application/json",
    )


def test_sagemaker_endpoint_inference_component_invoke() -> None:
    client = Mock()
    response = {
        "ContentType": "application/json",
        "Body": b'[{"generated_text": "SageMaker Endpoint"}]',
    }
    client.invoke_endpoint.return_value = response

    llm = SagemakerEndpoint(
        endpoint_name="my-endpoint",
        inference_component_name="my-inference-component",
        region_name="us-west-2",
        content_handler=DefaultHandler(),
        model_kwargs={
            "parameters": {
                "max_new_tokens": 50,
            }
        },
        client=client,
    )

    service_response = llm.invoke("What is Sagemaker endpoints?")

    assert service_response == "SageMaker Endpoint"
    client.invoke_endpoint.assert_called_once_with(
        EndpointName="my-endpoint",
        Body=b"What is Sagemaker endpoints?",
        ContentType="application/json",
        Accept="application/json",
        InferenceComponentName="my-inference-component",
    )


def test_sagemaker_endpoint_stream() -> None:
    class ContentHandler(LLMContentHandler):
        accepts = "application/json"
        content_type = "application/json"

        def transform_input(self, prompt: str, model_kwargs: Dict) -> bytes:
            body = json.dumps({"inputs": prompt, **model_kwargs})
            return body.encode()

        def transform_output(self, output: bytes) -> str:
            body = json.loads(output)
            return body.get("outputs")[0]

    body = (
        {"PayloadPart": {"Bytes": b'{"outputs": ["S"]}\n'}},
        {"PayloadPart": {"Bytes": b'{"outputs": ["age"]}\n'}},
        {"PayloadPart": {"Bytes": b'{"outputs": ["Maker"]}\n'}},
    )

    response = {"ContentType": "application/json", "Body": body}

    client = Mock()
    client.invoke_endpoint_with_response_stream.return_value = response

    llm = SagemakerEndpoint(
        endpoint_name="my-endpoint",
        region_name="us-west-2",
        content_handler=ContentHandler(),
        client=client,
        model_kwargs={"parameters": {"max_new_tokens": 50}},
    )

    expected_body = json.dumps(
        {"inputs": "What is Sagemaker endpoints?", "parameters": {"max_new_tokens": 50}}
    ).encode()

    chunks = ["S", "age", "Maker"]
    service_chunks = []

    for chunk in llm.stream("What is Sagemaker endpoints?"):
        service_chunks.append(chunk)

    assert service_chunks == chunks
    client.invoke_endpoint_with_response_stream.assert_called_once_with(
        EndpointName="my-endpoint",
        Body=expected_body,
        ContentType="application/json",
    )


def test_sagemaker_endpoint_inference_component_stream() -> None:
    class ContentHandler(LLMContentHandler):
        accepts = "application/json"
        content_type = "application/json"

        def transform_input(self, prompt: str, model_kwargs: Dict) -> bytes:
            body = json.dumps({"inputs": prompt, **model_kwargs})
            return body.encode()

        def transform_output(self, output: bytes) -> str:
            body = json.loads(output)
            return body.get("outputs")[0]

    body = (
        {"PayloadPart": {"Bytes": b'{"outputs": ["S"]}\n'}},
        {"PayloadPart": {"Bytes": b'{"outputs": ["age"]}\n'}},
        {"PayloadPart": {"Bytes": b'{"outputs": ["Maker"]}\n'}},
    )

    response = {"ContentType": "application/json", "Body": body}

    client = Mock()
    client.invoke_endpoint_with_response_stream.return_value = response

    llm = SagemakerEndpoint(
        endpoint_name="my-endpoint",
        inference_component_name="my_inference_component",
        region_name="us-west-2",
        content_handler=ContentHandler(),
        client=client,
        model_kwargs={"parameters": {"max_new_tokens": 50}},
    )

    expected_body = json.dumps(
        {"inputs": "What is Sagemaker endpoints?", "parameters": {"max_new_tokens": 50}}
    ).encode()

    chunks = ["S", "age", "Maker"]
    service_chunks = []

    for chunk in llm.stream("What is Sagemaker endpoints?"):
        service_chunks.append(chunk)

    assert service_chunks == chunks
    client.invoke_endpoint_with_response_stream.assert_called_once_with(
        EndpointName="my-endpoint",
        Body=expected_body,
        ContentType="application/json",
        InferenceComponentName="my_inference_component",
    )



from unittest.mock import Mock

import pytest
from langchain_core.documents import Document

from langchain_aws import AmazonKnowledgeBasesRetriever


@pytest.fixture
def mock_client() -> Mock:
    return Mock()


@pytest.fixture
def retriever(mock_client: Mock) -> AmazonKnowledgeBasesRetriever:
    return AmazonKnowledgeBasesRetriever(
        knowledge_base_id="test-knowledge-base",
        client=mock_client,
        retrieval_config={"vectorSearchConfiguration": {"numberOfResults": 4}},  # type: ignore[arg-type]
        min_score_confidence=0.0,
    )


def test_get_relevant_documents(retriever, mock_client) -> None:  # type: ignore[no-untyped-def]
    response = {
        "retrievalResults": [
            {
                "content": {"text": "This is the first result."},
                "location": "location1",
                "score": 0.9,
            },
            {
                "content": {"text": "This is the second result."},
                "location": "location2",
                "score": 0.8,
            },
            {"content": {"text": "This is the third result."}, "location": "location3"},
            {
                "content": {"text": "This is the fourth result."},
                "metadata": {"key1": "value1", "key2": "value2"},
            },
        ]
    }
    mock_client.retrieve.return_value = response

    query = "test query"

    expected_documents = [
        Document(
            page_content="This is the first result.",
            metadata={"location": "location1", "score": 0.9, "type": "TEXT"},
        ),
        Document(
            page_content="This is the second result.",
            metadata={"location": "location2", "score": 0.8, "type": "TEXT"},
        ),
        Document(
            page_content="This is the third result.",
            metadata={"location": "location3", "score": 0.0, "type": "TEXT"},
        ),
        Document(
            page_content="This is the fourth result.",
            metadata={
                "type": "TEXT",
                "score": 0.0,
                "source_metadata": {
                    "key1": "value1",
                    "key2": "value2",
                },
            },
        ),
    ]

    documents = retriever.invoke(query)

    assert documents == expected_documents

    mock_client.retrieve.assert_called_once_with(
        retrievalQuery={"text": "test query"},
        knowledgeBaseId="test-knowledge-base",
        retrievalConfiguration={"vectorSearchConfiguration": {"numberOfResults": 4}},
    )


def test_get_relevant_documents_with_score(retriever, mock_client) -> None:  # type: ignore[no-untyped-def]
    response = {
        "retrievalResults": [
            {
                "content": {"text": "This is the first result."},
                "location": "location1",
                "score": 0.9,
            },
            {
                "content": {"text": "This is the second result."},
                "location": "location2",
                "score": 0.8,
            },
            {"content": {"text": "This is the third result."}, "location": "location3"},
            {
                "content": {"text": "This is the fourth result."},
                "metadata": {"key1": "value1", "key2": "value2"},
            },
        ]
    }
    mock_client.retrieve.return_value = response

    query = "test query"

    expected_documents = [
        Document(
            page_content="This is the first result.",
            metadata={"location": "location1", "score": 0.9, "type": "TEXT"},
        ),
        Document(
            page_content="This is the second result.",
            metadata={"location": "location2", "score": 0.8, "type": "TEXT"},
        ),
    ]

    retriever.min_score_confidence = 0.80
    documents = retriever.invoke(query)

    assert documents == expected_documents



"""Integration tests for Nova Lite 2.0 with system tools and reasoning.

This module consolidates all Nova tools integration tests including:
- Nova Grounding (web search)
- Nova Code Interpreter
- Reasoning configuration
- Streaming behavior
"""

from typing import Any, Generator

import pytest
from langchain_core.messages import AIMessage, AIMessageChunk, HumanMessage

from langchain_aws import ChatBedrockConverse
from langchain_aws.tools import (
    NovaCodeInterpreterTool,
    NovaGroundingTool,
)

MODEL_ID = "us.amazon.nova-2-lite-v1:0"


# =============================================================================
# Fixtures
# =============================================================================


@pytest.fixture
def nova_model() -> Generator[ChatBedrockConverse, None, None]:
    """Create a ChatBedrockConverse model for Nova 2 Lite."""
    yield ChatBedrockConverse(
        model=MODEL_ID,
        max_tokens=10000,
    )


@pytest.fixture
def nova_model_reasoning_low() -> Generator[ChatBedrockConverse, None, None]:
    """Create a ChatBedrockConverse model for Nova 2 Lite with LOW reasoning."""
    yield ChatBedrockConverse(
        model=MODEL_ID,
        max_tokens=10000,
        additional_model_request_fields={
            "reasoningConfig": {
                "type": "enabled",
                "maxReasoningEffort": "low",
            }
        },
    )


@pytest.fixture
def nova_model_reasoning_medium() -> Generator[ChatBedrockConverse, None, None]:
    """Create a ChatBedrockConverse model for Nova 2 Lite with MEDIUM reasoning."""
    yield ChatBedrockConverse(
        model=MODEL_ID,
        max_tokens=10000,
        additional_model_request_fields={
            "reasoningConfig": {
                "type": "enabled",
                "maxReasoningEffort": "medium",
            }
        },
    )


@pytest.fixture
def nova_model_reasoning_high() -> Generator[ChatBedrockConverse, None, None]:
    """Create a ChatBedrockConverse model for Nova 2 Lite with HIGH reasoning."""
    yield ChatBedrockConverse(
        model=MODEL_ID,
        max_tokens=10000,
        additional_model_request_fields={
            "reasoningConfig": {
                "type": "enabled",
                "maxReasoningEffort": "high",
            }
        },
    )


@pytest.fixture
def nova_model_reasoning_disabled() -> Generator[ChatBedrockConverse, None, None]:
    """Create a ChatBedrockConverse model for Nova 2 Lite with reasoning disabled."""
    yield ChatBedrockConverse(
        model=MODEL_ID,
        max_tokens=10000,
        additional_model_request_fields={
            "reasoningConfig": {
                "type": "disabled",
            }
        },
    )


# =============================================================================
# Nova Grounding Tool Tests
# =============================================================================


class TestNovaGroundingTool:
    """Tests for nova_grounding system tool (web search)."""

    def test_basic(self, nova_model: ChatBedrockConverse) -> None:
        """Test nova_grounding system tool with real API call."""
        model_with_tools = nova_model.bind_tools([NovaGroundingTool()])
        response = model_with_tools.invoke("Who won the 2024 Nobel Prize in Physics?")

        # Verify response structure
        assert isinstance(response, AIMessage)
        assert len(response.content) > 0

        # Verify the response contains text
        assert any(
            isinstance(block, dict) and block.get("type") == "text"
            for block in response.content
        )

    def test_with_string(self, nova_model: ChatBedrockConverse) -> None:
        """Test nova_grounding using direct string instead of helper class."""
        model_with_tools = nova_model.bind_tools(["nova_grounding"])
        response = model_with_tools.invoke("What is the current weather in Seattle?")

        assert isinstance(response, AIMessage)
        assert len(response.content) > 0

    def test_with_reasoning_enabled(
        self, nova_model_reasoning_low: ChatBedrockConverse
    ) -> None:
        """Test nova_grounding with reasoning enabled."""
        model_with_tools = nova_model_reasoning_low.bind_tools([NovaGroundingTool()])
        response = model_with_tools.invoke(
            "Who won the Oscar for best actress in 2024?"
        )

        # Verify response structure
        assert isinstance(response, AIMessage)
        assert len(response.content) > 0

        # Verify content blocks include reasoning
        content_blocks = response.content_blocks
        assert any(block["type"] == "reasoning" for block in content_blocks)

    def test_content_blocks(
        self, nova_model_reasoning_low: ChatBedrockConverse
    ) -> None:
        """Test that content blocks are properly parsed for nova_grounding."""
        model_with_tools = nova_model_reasoning_low.bind_tools([NovaGroundingTool()])
        response = model_with_tools.invoke(
            "What are the latest developments in quantum computing?"
        )

        # Verify response structure
        assert isinstance(response, AIMessage)

        # Get content blocks
        content_blocks = response.content_blocks
        assert len(content_blocks) > 0

        # Verify we have reasoning content
        reasoning_blocks = [b for b in content_blocks if b["type"] == "reasoning"]
        assert len(reasoning_blocks) > 0

        # Verify we have text content
        text_blocks = [b for b in content_blocks if b["type"] == "text"]
        assert len(text_blocks) > 0

    def test_streaming(self, nova_model_reasoning_low: ChatBedrockConverse) -> None:
        """Test streaming with nova_grounding system tool."""
        model_with_tools = nova_model_reasoning_low.bind_tools([NovaGroundingTool()])

        chunks = []
        for chunk in model_with_tools.stream("What's the latest news about AI?"):
            chunks.append(chunk)

        # Verify we received chunks
        assert len(chunks) > 0

        # Verify final message by accumulating chunks
        full_message: Any = chunks[0]
        for chunk in chunks[1:]:
            full_message += chunk

        assert isinstance(full_message, AIMessage)
        assert len(full_message.content) > 0

    def test_with_different_reasoning_levels(self) -> None:
        """Test nova_grounding with different reasoning effort levels."""
        for effort in ["low", "medium", "high"]:
            model = ChatBedrockConverse(
                model=MODEL_ID,
                max_tokens=10000,
                additional_model_request_fields={
                    "reasoningConfig": {
                        "type": "enabled",
                        "maxReasoningEffort": effort,
                    }
                },
            )

            model_with_tools = model.bind_tools([NovaGroundingTool()])
            response = model_with_tools.invoke("What is the capital of France?")

            assert isinstance(response, AIMessage)
            assert len(response.content) > 0

            # Verify reasoning content exists
            content_blocks = response.content_blocks
            reasoning_blocks = [b for b in content_blocks if b["type"] == "reasoning"]
            assert len(reasoning_blocks) > 0

    def test_multi_turn_conversation(self, nova_model: ChatBedrockConverse) -> None:
        """Test nova_grounding in a multi-turn conversation."""
        model_with_tools = nova_model.bind_tools([NovaGroundingTool()])

        # First turn
        message1 = HumanMessage("What's the population of Tokyo?")
        response1 = model_with_tools.invoke([message1])
        assert isinstance(response1, AIMessage)

        # Second turn - follow-up question
        message2 = HumanMessage("How does that compare to New York City?")
        response2 = model_with_tools.invoke([message1, response1, message2])
        assert isinstance(response2, AIMessage)
        assert len(response2.content) > 0


# =============================================================================
# Nova Code Interpreter Tool Tests
# =============================================================================


class TestNovaCodeInterpreterTool:
    """Tests for nova_code_interpreter system tool."""

    def test_basic(self, nova_model: ChatBedrockConverse) -> None:
        """Test nova_code_interpreter system tool with real API call."""
        model_with_tools = nova_model.bind_tools([NovaCodeInterpreterTool()])
        response = model_with_tools.invoke("Use Python to calculate 123 * 456")

        # Verify response structure
        assert isinstance(response, AIMessage)
        assert len(response.content) > 0

        # Verify the response contains text with the result (123 * 456 = 56088)
        response_text = str(response.content)
        assert "56088" in response_text or "56,088" in response_text

    def test_with_string(self, nova_model: ChatBedrockConverse) -> None:
        """Test nova_code_interpreter using direct string instead of helper class."""
        model_with_tools = nova_model.bind_tools(["nova_code_interpreter"])
        response = model_with_tools.invoke("What is 123 * 456?")

        assert isinstance(response, AIMessage)
        assert len(response.content) > 0

        # Verify the calculation result (123 * 456 = 56088)
        response_text = str(response.content)
        assert "56088" in response_text or "56,088" in response_text

    def test_simple_calculation(self, nova_model: ChatBedrockConverse) -> None:
        """Test nova_code_interpreter with simple calculation."""
        model_with_tools = nova_model.bind_tools([NovaCodeInterpreterTool()])
        response = model_with_tools.invoke("Use Python code to calculate 7 ** 6")

        # Verify response structure
        assert isinstance(response, AIMessage)
        assert len(response.content) > 0

        # Verify the calculation result (7^6 = 117649)
        response_text = str(response.content)
        assert "117649" in response_text or "117,649" in response_text

    def test_with_reasoning(
        self, nova_model_reasoning_low: ChatBedrockConverse
    ) -> None:
        """Test nova_code_interpreter with reasoning enabled."""
        model_with_tools = nova_model_reasoning_low.bind_tools(
            [NovaCodeInterpreterTool()]
        )
        response = model_with_tools.invoke(
            "Calculate the factorial of 10 and explain the result. "
            "Use Code Interpreter to calculate it."
        )

        # Verify response structure
        assert isinstance(response, AIMessage)
        assert len(response.content) > 0

        # Verify content blocks include reasoning
        content_blocks = response.content_blocks
        assert any(block["type"] == "reasoning" for block in content_blocks)

        # Verify the calculation result (10! = 3628800)
        response_text = str(response.content)
        assert "3628800" in response_text or "3,628,800" in response_text

    def test_content_blocks(
        self, nova_model_reasoning_low: ChatBedrockConverse
    ) -> None:
        """Test that content blocks are properly parsed for nova_code_interpreter."""
        model_with_tools = nova_model_reasoning_low.bind_tools(
            [NovaCodeInterpreterTool()]
        )
        response = model_with_tools.invoke("Calculate the sum of numbers from 1 to 100")

        # Verify response structure
        assert isinstance(response, AIMessage)

        # Get content blocks
        content_blocks = response.content_blocks
        assert len(content_blocks) > 0

        # Verify we have reasoning content
        reasoning_blocks = [b for b in content_blocks if b["type"] == "reasoning"]
        assert len(reasoning_blocks) > 0

        # Verify we have text content
        text_blocks = [b for b in content_blocks if b["type"] == "text"]
        assert len(text_blocks) > 0

        # Verify the calculation result (sum 1-100 = 5050)
        response_text = str(response.content)
        assert "5050" in response_text or "5,050" in response_text

    def test_streaming(self, nova_model: ChatBedrockConverse) -> None:
        """Test streaming with nova_code_interpreter system tool."""
        model_with_tools = nova_model.bind_tools([NovaCodeInterpreterTool()])

        chunks = []
        for chunk in model_with_tools.stream("What is 999 * 888?"):
            chunks.append(chunk)

        # Verify we received chunks
        assert len(chunks) > 0

        # Verify final message by accumulating chunks
        full_message: Any = chunks[0]
        for chunk in chunks[1:]:
            full_message += chunk

        assert isinstance(full_message, AIMessage)
        assert len(full_message.content) > 0

        # Verify the calculation result (999 * 888 = 887112)
        response_text = str(full_message.content)
        assert "887112" in response_text or "887,112" in response_text

    def test_multi_turn_conversation(self, nova_model: ChatBedrockConverse) -> None:
        """Test nova_code_interpreter in a multi-turn conversation."""
        model_with_tools = nova_model.bind_tools([NovaCodeInterpreterTool()])

        # First turn
        message1 = HumanMessage("Calculate 15 * 20")
        response1 = model_with_tools.invoke([message1])
        assert isinstance(response1, AIMessage)
        assert "300" in str(response1.content)

        # Second turn - follow-up calculation
        message2 = HumanMessage("Now add 50 to that result")
        response2 = model_with_tools.invoke([message1, response1, message2])
        assert isinstance(response2, AIMessage)
        assert len(response2.content) > 0
        assert "350" in str(response2.content)


# =============================================================================
# Reasoning Configuration Tests
# =============================================================================


class TestNovaReasoning:
    """Tests for Nova reasoning configuration levels."""

    def test_effort_low(self, nova_model_reasoning_low: ChatBedrockConverse) -> None:
        """Test reasoning with effort level 'low'."""
        response = nova_model_reasoning_low.invoke("What is 7 to the power of 6?")

        # Verify response structure
        assert isinstance(response, AIMessage)
        assert len(response.content) > 0

        # Verify reasoning content blocks are present
        content_blocks = response.content_blocks
        reasoning_blocks = [b for b in content_blocks if b["type"] == "reasoning"]
        assert len(reasoning_blocks) > 0, (
            "Expected reasoning content blocks with LOW effort"
        )

        # Verify reasoning content has text
        for block in reasoning_blocks:
            assert "reasoning" in block
            assert len(block["reasoning"]) > 0

    def test_effort_medium(
        self, nova_model_reasoning_medium: ChatBedrockConverse
    ) -> None:
        """Test reasoning with effort level 'medium'."""
        response = nova_model_reasoning_medium.invoke(
            "Explain the concept of quantum entanglement"
        )

        # Verify response structure
        assert isinstance(response, AIMessage)
        assert len(response.content) > 0

        # Verify reasoning content blocks are present
        content_blocks = response.content_blocks
        reasoning_blocks = [b for b in content_blocks if b["type"] == "reasoning"]
        assert len(reasoning_blocks) > 0, (
            "Expected reasoning content blocks with MEDIUM effort"
        )

        # Verify reasoning content has text
        for block in reasoning_blocks:
            assert "reasoning" in block
            assert len(block["reasoning"]) > 0

    def test_effort_high(self, nova_model_reasoning_high: ChatBedrockConverse) -> None:
        """Test reasoning with effort level 'high'."""
        response = nova_model_reasoning_high.invoke(
            "What are the implications of the halting problem?"
        )

        # Verify response structure
        assert isinstance(response, AIMessage)
        assert len(response.content) > 0

        # Verify reasoning content blocks are present
        content_blocks = response.content_blocks
        reasoning_blocks = [b for b in content_blocks if b["type"] == "reasoning"]
        assert len(reasoning_blocks) > 0, (
            "Expected reasoning content blocks with HIGH effort"
        )

        # Verify reasoning content has text
        for block in reasoning_blocks:
            assert "reasoning" in block
            assert len(block["reasoning"]) > 0

    def test_effort_string_values(self) -> None:
        """Test reasoning with string effort levels instead of enum."""
        for effort in ["low", "medium", "high"]:
            model = ChatBedrockConverse(
                model=MODEL_ID,
                max_tokens=10000,
                additional_model_request_fields={
                    "reasoningConfig": {
                        "type": "enabled",
                        "maxReasoningEffort": effort,
                    }
                },
            )

            response = model.invoke(f"Calculate {effort}: What is 12 * 13?")

            # Verify response structure
            assert isinstance(response, AIMessage)
            assert len(response.content) > 0

            # Verify reasoning content blocks are present
            content_blocks = response.content_blocks
            reasoning_blocks = [b for b in content_blocks if b["type"] == "reasoning"]
            assert len(reasoning_blocks) > 0, (
                f"Expected reasoning content blocks with '{effort}' effort"
            )

    def test_disabled(self, nova_model_reasoning_disabled: ChatBedrockConverse) -> None:
        """Test that reasoning blocks are not present when reasoning is disabled."""
        response = nova_model_reasoning_disabled.invoke(
            "What is the capital of France?"
        )

        # Verify response structure
        assert isinstance(response, AIMessage)
        assert len(response.content) > 0

        # Verify no reasoning content blocks are present
        content_blocks = response.content_blocks
        reasoning_blocks = [b for b in content_blocks if b["type"] == "reasoning"]
        assert len(reasoning_blocks) == 0, (
            "Expected no reasoning content blocks when disabled"
        )

    def test_default_no_config(self, nova_model: ChatBedrockConverse) -> None:
        """Test that reasoning is disabled by default when no config is provided."""
        response = nova_model.invoke("What is 5 + 5?")

        # Verify response structure
        assert isinstance(response, AIMessage)
        assert len(response.content) > 0

        # Verify no reasoning content blocks are present (default is disabled)
        content_blocks = response.content_blocks
        [b for b in content_blocks if b["type"] == "reasoning"]
        # Note: Some models may still include reasoning even without explicit config
        # So we just verify the response is valid, not that reasoning is absent

    def test_content_structure(
        self, nova_model_reasoning_low: ChatBedrockConverse
    ) -> None:
        """Test the structure of reasoning content blocks."""
        response = nova_model_reasoning_low.invoke(
            "Solve this: If x + 5 = 12, what is x?"
        )

        # Verify response structure
        assert isinstance(response, AIMessage)

        # Get content blocks
        content_blocks = response.content_blocks
        assert len(content_blocks) > 0

        # Find reasoning blocks
        reasoning_blocks = [b for b in content_blocks if b["type"] == "reasoning"]
        assert len(reasoning_blocks) > 0

        # Verify reasoning block structure
        for block in reasoning_blocks:
            assert isinstance(block, dict)
            assert "type" in block
            assert block["type"] == "reasoning"
            assert "reasoning" in block
            assert isinstance(block["reasoning"], str)
            assert len(block["reasoning"]) > 0

    def test_with_complex_query(
        self, nova_model_reasoning_medium: ChatBedrockConverse
    ) -> None:
        """Test reasoning with a complex query that requires deeper thinking."""
        response = nova_model_reasoning_medium.invoke(
            "If a train travels at 60 mph for 2 hours, then 80 mph for 1.5 hours, "
            "what is the total distance traveled?"
        )

        # Verify response structure
        assert isinstance(response, AIMessage)
        assert len(response.content) > 0

        # Verify reasoning content blocks are present
        content_blocks = response.content_blocks
        reasoning_blocks = [b for b in content_blocks if b["type"] == "reasoning"]
        assert len(reasoning_blocks) > 0

        # Verify the response contains the calculation
        response_text = str(response.content).lower()
        # Total distance = (60 * 2) + (80 * 1.5) = 120 + 120 = 240 miles
        assert "240" in response_text or "distance" in response_text


# =============================================================================
# Streaming Tests
# =============================================================================


class TestNovaStreaming:
    """Tests for Nova streaming behavior with system tools."""

    def test_with_grounding(self, nova_model: ChatBedrockConverse) -> None:
        """Test streaming with nova_grounding system tool."""
        model_with_tools = nova_model.bind_tools([NovaGroundingTool()])

        chunks = []
        for chunk in model_with_tools.stream("What's the latest news about SpaceX?"):
            chunks.append(chunk)
            # Verify each chunk is an AIMessageChunk
            assert isinstance(chunk, AIMessageChunk)

        # Verify we received chunks
        assert len(chunks) > 0

        # Verify final message by accumulating chunks
        full_message: Any = chunks[0]
        for chunk in chunks[1:]:
            full_message += chunk

        assert isinstance(full_message, AIMessage)
        assert len(full_message.content) > 0

    def test_with_code_interpreter(self, nova_model: ChatBedrockConverse) -> None:
        """Test streaming with nova_code_interpreter system tool."""
        model_with_tools = nova_model.bind_tools([NovaCodeInterpreterTool()])

        chunks = []
        for chunk in model_with_tools.stream("Calculate 456 * 789 using Python"):
            chunks.append(chunk)
            # Verify each chunk is an AIMessageChunk
            assert isinstance(chunk, AIMessageChunk)

        # Verify we received chunks
        assert len(chunks) > 0

        # Verify final message by accumulating chunks
        full_message: Any = chunks[0]
        for chunk in chunks[1:]:
            full_message += chunk

        assert isinstance(full_message, AIMessage)
        assert len(full_message.content) > 0

        # Verify the calculation result (456 * 789 = 359784)
        response_text = str(full_message.content)
        assert "359784" in response_text or "359,784" in response_text

    def test_reasoning_content_chunks(
        self, nova_model_reasoning_low: ChatBedrockConverse
    ) -> None:
        """Test that reasoning content chunks stream correctly."""
        model_with_tools = nova_model_reasoning_low.bind_tools([NovaGroundingTool()])

        chunks = []
        reasoning_chunks = []

        for chunk in model_with_tools.stream("What are the latest AI breakthroughs?"):
            chunks.append(chunk)

            # Check if this chunk contains reasoning content
            if hasattr(chunk, "content") and isinstance(chunk.content, list):
                for block in chunk.content:
                    if isinstance(block, dict) and block.get("type") == "reasoning":
                        reasoning_chunks.append(block)

        # Verify we received chunks
        assert len(chunks) > 0

        # Accumulate full message
        full_message: Any = chunks[0]
        for chunk in chunks[1:]:
            full_message += chunk

        assert isinstance(full_message, AIMessage)

        # Verify reasoning content exists in the full message
        content_blocks = full_message.content_blocks
        reasoning_blocks = [b for b in content_blocks if b["type"] == "reasoning"]
        assert len(reasoning_blocks) > 0, (
            "Expected reasoning content blocks in streamed message"
        )

    def test_tool_use_and_result_events(self, nova_model: ChatBedrockConverse) -> None:
        """Test that tool use and tool result events stream correctly."""
        model_with_tools = nova_model.bind_tools([NovaCodeInterpreterTool()])

        chunks = []
        tool_use_chunks = []

        for chunk in model_with_tools.stream(
            "Use Python to calculate the square root of 144"
        ):
            chunks.append(chunk)

            # Check if this chunk contains tool use content
            if hasattr(chunk, "content") and isinstance(chunk.content, list):
                for block in chunk.content:
                    if isinstance(block, dict) and block.get("type") == "tool_use":
                        tool_use_chunks.append(block)

        # Verify we received chunks
        assert len(chunks) > 0

        # Accumulate full message
        full_message: Any = chunks[0]
        for chunk in chunks[1:]:
            full_message += chunk

        assert isinstance(full_message, AIMessage)
        assert len(full_message.content) > 0

        # Verify the calculation result (sqrt(144) = 12)
        response_text = str(full_message.content)
        assert "12" in response_text

    def test_accumulating_chunks_into_full_message(
        self, nova_model_reasoning_low: ChatBedrockConverse
    ) -> None:
        """Test accumulating chunks into full message with all content types."""
        model_with_tools = nova_model_reasoning_low.bind_tools([NovaGroundingTool()])

        chunks = []
        for chunk in model_with_tools.stream(
            "What is the current population of Tokyo?"
        ):
            chunks.append(chunk)
            assert isinstance(chunk, AIMessageChunk)

        # Verify we received chunks
        assert len(chunks) > 0

        # Accumulate chunks into full message
        full_message: Any = chunks[0]
        for chunk in chunks[1:]:
            full_message += chunk

        # Verify full message structure
        assert isinstance(full_message, AIMessage)
        assert len(full_message.content) > 0

        # Verify content blocks are properly accumulated
        content_blocks = full_message.content_blocks
        assert len(content_blocks) > 0

        # Verify we have reasoning blocks
        reasoning_blocks = [b for b in content_blocks if b["type"] == "reasoning"]
        assert len(reasoning_blocks) > 0

        # Verify we have text blocks
        text_blocks = [b for b in content_blocks if b["type"] == "text"]
        assert len(text_blocks) > 0

    def test_with_both_system_tools(self, nova_model: ChatBedrockConverse) -> None:
        """Test streaming with both nova_grounding and nova_code_interpreter."""
        model_with_tools = nova_model.bind_tools(
            [
                NovaGroundingTool(),
                NovaCodeInterpreterTool(),
            ]
        )

        chunks = []
        for chunk in model_with_tools.stream(
            "Search for the current price of Bitcoin and calculate 10% of that value"
        ):
            chunks.append(chunk)
            assert isinstance(chunk, AIMessageChunk)

        # Verify we received chunks
        assert len(chunks) > 0

        # Accumulate full message
        full_message: Any = chunks[0]
        for chunk in chunks[1:]:
            full_message += chunk

        assert isinstance(full_message, AIMessage)
        assert len(full_message.content) > 0

    def test_reasoning_with_different_effort_levels(self) -> None:
        """Test streaming with different reasoning effort levels."""
        for effort in ["low", "medium", "high"]:
            model = ChatBedrockConverse(
                model=MODEL_ID,
                max_tokens=10000,
                additional_model_request_fields={
                    "reasoningConfig": {
                        "type": "enabled",
                        "maxReasoningEffort": effort,
                    }
                },
            )

            model_with_tools = model.bind_tools([NovaCodeInterpreterTool()])

            chunks = []
            for chunk in model_with_tools.stream("Calculate 15 factorial"):
                chunks.append(chunk)

            # Verify we received chunks
            assert len(chunks) > 0

            # Accumulate full message
            full_message: Any = chunks[0]
            for chunk in chunks[1:]:
                full_message += chunk

            assert isinstance(full_message, AIMessage)

            # Verify reasoning content exists
            content_blocks = full_message.content_blocks
            reasoning_blocks = [b for b in content_blocks if b["type"] == "reasoning"]
            assert len(reasoning_blocks) > 0, (
                f"Expected reasoning blocks with {effort} effort"
            )

    def test_chunk_content_structure(
        self, nova_model_reasoning_low: ChatBedrockConverse
    ) -> None:
        """Test the structure of streaming chunks."""
        model_with_tools = nova_model_reasoning_low.bind_tools([NovaGroundingTool()])

        chunks = []
        for chunk in model_with_tools.stream("What is the weather in Paris?"):
            chunks.append(chunk)

            # Verify chunk structure
            assert isinstance(chunk, AIMessageChunk)
            assert hasattr(chunk, "content")

            # Content should be a list or string
            assert isinstance(chunk.content, (list, str))

        # Verify we received chunks
        assert len(chunks) > 0

    def test_with_string_tool_names(self, nova_model: ChatBedrockConverse) -> None:
        """Test streaming with system tools specified as strings."""
        model_with_tools = nova_model.bind_tools(
            ["nova_grounding", "nova_code_interpreter"]
        )

        chunks = []
        for chunk in model_with_tools.stream("What is 25 * 25?"):
            chunks.append(chunk)
            assert isinstance(chunk, AIMessageChunk)

        # Verify we received chunks
        assert len(chunks) > 0

        # Accumulate full message
        full_message: Any = chunks[0]
        for chunk in chunks[1:]:
            full_message += chunk

        assert isinstance(full_message, AIMessage)
        assert len(full_message.content) > 0

        # Verify the calculation result (25 * 25 = 625)
        response_text = str(full_message.content)
        assert "625" in response_text

    def test_empty_chunks_handling(self, nova_model: ChatBedrockConverse) -> None:
        """Test that empty or minimal chunks are handled correctly."""
        model_with_tools = nova_model.bind_tools([NovaCodeInterpreterTool()])

        chunks = []
        for chunk in model_with_tools.stream("What is 2 + 2?"):
            chunks.append(chunk)
            # Each chunk should be valid even if content is minimal
            assert isinstance(chunk, AIMessageChunk)

        # Verify we received chunks
        assert len(chunks) > 0

        # Accumulate full message
        full_message: Any = chunks[0]
        for chunk in chunks[1:]:
            full_message += chunk

        assert isinstance(full_message, AIMessage)
        # Even for simple queries, we should get a response
        assert len(full_message.content) > 0



"""Bedrock Anthropic prompt caching middleware.

Requires:
    - langchain: For agent middleware framework
    - langchain-aws: For ChatBedrock model (already a dependency)
"""

from collections.abc import Awaitable, Callable
from typing import Literal
from warnings import warn

from langchain_aws.chat_models.bedrock import ChatBedrock

try:
    from langchain.agents.middleware.types import (
        AgentMiddleware,
        ModelCallResult,
        ModelRequest,
        ModelResponse,
    )
except ImportError as e:
    msg = (
        "Prompt caching middleware requires 'langchain' to be installed. "
        "This middleware is designed for use with LangChain agents. "
        "Install it with: pip install langchain"
    )
    raise ImportError(msg) from e


def _is_anthropic_model(model: ChatBedrock) -> bool:
    """Check if the model is an Anthropic model on Bedrock."""
    model_id = getattr(model, "model_id", "") or getattr(model, "model", "")
    return "anthropic" in model_id.lower()


class BedrockPromptCachingMiddleware(AgentMiddleware):
    """Prompt Caching Middleware for ChatBedrock (InvokeModel API).

    Optimizes API usage by caching conversation prefixes for Anthropic models
    on AWS Bedrock. Adds cache_control to the last message, caching the entire
    conversation prefix (including system prompt) for improved cache hits in
    multi-turn conversations.

    Requires both 'langchain' and 'langchain-aws' packages to be installed.

    Learn more about Anthropic prompt caching
    [here](https://docs.anthropic.com/en/docs/build-with-claude/prompt-caching).
    """

    def __init__(
        self,
        type: Literal["ephemeral"] = "ephemeral",  # noqa: A002
        ttl: Literal["5m", "1h"] = "5m",
        min_messages_to_cache: int = 0,
        unsupported_model_behavior: Literal["ignore", "warn", "raise"] = "warn",
    ) -> None:
        """Initialize the middleware with cache control settings.

        Args:
            type: The type of cache to use, only "ephemeral" is supported.
            ttl: The time to live for the cache, only "5m" and "1h" are
                supported.
            min_messages_to_cache: The minimum number of messages until the
                cache is used, default is 0.
            unsupported_model_behavior: The behavior to take when an
                unsupported model is used. "ignore" will ignore the unsupported
                model and continue without caching. "warn" will warn the user
                and continue without caching. "raise" will raise an error and
                stop the agent.
        """
        self.type = type
        self.ttl = ttl
        self.min_messages_to_cache = min_messages_to_cache
        self.unsupported_model_behavior = unsupported_model_behavior

    def _should_apply_caching(self, request: ModelRequest) -> bool:
        """Check if caching should be applied to the request."""
        if not isinstance(request.model, ChatBedrock):
            msg = (
                f"BedrockPromptCachingMiddleware only supports ChatBedrock, "
                f"not {type(request.model).__name__}. "
                f"Converse API prompt caching support is planned for a future release."
            )
            if self.unsupported_model_behavior == "raise":
                raise ValueError(msg)
            if self.unsupported_model_behavior == "warn":
                warn(msg, stacklevel=3)
            return False

        if not _is_anthropic_model(request.model):
            model_id = getattr(request.model, "model_id", "unknown")
            msg = f"Prompt caching only supported for Anthropic models: {model_id}"
            if self.unsupported_model_behavior == "raise":
                raise ValueError(msg)
            if self.unsupported_model_behavior == "warn":
                warn(msg, stacklevel=3)
            return False

        messages_count = (
            len(request.messages) + 1
            if request.system_prompt
            else len(request.messages)
        )
        return messages_count >= self.min_messages_to_cache

    def _get_cache_control_settings(self) -> dict:
        """Get cache control settings to pass via model_settings."""
        return {"type": self.type, "ttl": self.ttl}

    def wrap_model_call(
        self,
        request: ModelRequest,
        handler: Callable[[ModelRequest], ModelResponse],
    ) -> ModelCallResult:
        """Wrap the model call to add cache control via model_settings."""
        if self._should_apply_caching(request):
            # Pass cache_control through model_settings instead of modifying messages.
            # This prevents cache_control from accumulating in checkpoints across turns.
            # ChatBedrock will apply this at API call time.
            new_model_settings = {
                **request.model_settings,
                "cache_control": self._get_cache_control_settings(),
            }
            return handler(request.override(model_settings=new_model_settings))
        return handler(request)

    async def awrap_model_call(
        self,
        request: ModelRequest,
        handler: Callable[[ModelRequest], Awaitable[ModelResponse]],
    ) -> ModelCallResult:
        """Wrap the model call to add cache control via model_settings (async)."""
        if self._should_apply_caching(request):
            # Pass cache_control through model_settings instead of modifying messages.
            # This prevents cache_control from accumulating in checkpoints across turns.
            # ChatBedrock will apply this at API call time.
            new_model_settings = {
                **request.model_settings,
                "cache_control": self._get_cache_control_settings(),
            }
            return await handler(request.override(model_settings=new_model_settings))
        return await handler(request)



import json
import logging
import re
import warnings
from collections import defaultdict
from operator import itemgetter
from typing import (
    Any,
    Callable,
    Dict,
    Iterator,
    List,
    Literal,
    Optional,
    Sequence,
    Tuple,
    Type,
    Union,
    cast,
)

from langchain_core.callbacks import CallbackManagerForLLMRun
from langchain_core.exceptions import OutputParserException
from langchain_core.language_models import (
    BaseChatModel,
    LangSmithParams,
    LanguageModelInput,
    ModelProfile,
    ModelProfileRegistry,
)
from langchain_core.language_models.chat_models import generate_from_stream
from langchain_core.messages import (
    AIMessage,
    AIMessageChunk,
    BaseMessage,
    ChatMessage,
    HumanMessage,
    SystemMessage,
    is_data_content_block,
)
from langchain_core.messages import content as types
from langchain_core.messages.ai import UsageMetadata
from langchain_core.messages.tool import ToolCall, ToolMessage
from langchain_core.messages.utils import convert_to_openai_messages
from langchain_core.output_parsers import JsonOutputKeyToolsParser, PydanticToolsParser
from langchain_core.output_parsers.base import OutputParserLike
from langchain_core.outputs import ChatGeneration, ChatGenerationChunk, ChatResult
from langchain_core.runnables import Runnable, RunnableMap, RunnablePassthrough
from langchain_core.tools import BaseTool
from langchain_core.utils import get_pydantic_field_names
from langchain_core.utils.function_calling import convert_to_openai_tool
from langchain_core.utils.pydantic import TypeBaseModel, is_basemodel_subclass
from langchain_core.utils.utils import _build_model_kwargs
from pydantic import BaseModel, ConfigDict, Field, model_validator
from typing_extensions import Self

from langchain_aws.chat_models._compat import _convert_from_v1_to_anthropic
from langchain_aws.chat_models.bedrock_converse import ChatBedrockConverse
from langchain_aws.data._profiles import _PROFILES
from langchain_aws.function_calling import (
    AnthropicTool,
    ToolsOutputParser,
    _lc_tool_calls_to_anthropic_tool_use_blocks,
    convert_to_anthropic_tool,
    get_system_message,
)
from langchain_aws.llms.bedrock import (
    BedrockBase,
    _citations_enabled,
    _combine_generation_info_for_llm_result,
)
from langchain_aws.utils import (
    anthropic_tokens_supported,
    count_tokens_api_supported_for_model,
    create_aws_client,
    get_num_tokens_anthropic,
    get_token_ids_anthropic,
    thinking_in_params,
    trim_message_whitespace,
)

logger = logging.getLogger(__name__)


_MODEL_PROFILES = cast("ModelProfileRegistry", _PROFILES)


def _get_default_model_profile(model_name: str) -> ModelProfile:
    default = _MODEL_PROFILES.get(model_name) or {}
    return default.copy()


def _convert_one_message_to_text_llama(message: BaseMessage) -> str:
    if isinstance(message, ChatMessage):
        message_text = f"\n\n{message.role.capitalize()}: {message.content}"
    elif isinstance(message, HumanMessage):
        message_text = f"[INST] {message.content} [/INST]"
    elif isinstance(message, AIMessage):
        message_text = f"{message.content}"
    elif isinstance(message, SystemMessage):
        message_text = f"<<SYS>> {message.content} <</SYS>>"
    else:
        raise ValueError(f"Got unknown type {message}")
    return message_text


def convert_messages_to_prompt_llama(messages: List[BaseMessage]) -> str:
    """Convert a list of messages to a prompt for llama."""

    return "\n".join(
        [_convert_one_message_to_text_llama(message) for message in messages]
    )


def _convert_one_message_to_text_llama3(message: BaseMessage) -> str:
    if isinstance(message, ChatMessage):
        message_text = (
            f"<|start_header_id|>{message.role}"
            f"<|end_header_id|>{message.content}<|eot_id|>"
        )
    elif isinstance(message, HumanMessage):
        message_text = (
            f"<|start_header_id|>user<|end_header_id|>{message.content}<|eot_id|>"
        )
    elif isinstance(message, AIMessage):
        message_text = (
            f"<|start_header_id|>assistant<|end_header_id|>{message.content}<|eot_id|>"
        )
    elif isinstance(message, SystemMessage):
        message_text = (
            f"<|start_header_id|>system<|end_header_id|>{message.content}<|eot_id|>"
        )
    else:
        raise ValueError(f"Got unknown type {message}")

    return message_text


def convert_messages_to_prompt_llama3(messages: List[BaseMessage]) -> str:
    """Convert a list of messages to a prompt for Llama 3."""

    return "\n".join(
        ["<|begin_of_text|>"]
        + [_convert_one_message_to_text_llama3(message) for message in messages]
        + ["<|start_header_id|>assistant<|end_header_id|>\n\n"]
    )


def _convert_one_message_to_text_llama4(message: BaseMessage) -> str:
    if isinstance(message, ChatMessage):
        message_text = (
            f"<|header_start|>{message.role}<|header_end|>{message.content}<|eot|>"
        )
    elif isinstance(message, HumanMessage):
        message_text = f"<|header_start|>user<|header_end|>{message.content}<|eot|>"
    elif isinstance(message, AIMessage):
        message_text = (
            f"<|header_start|>assistant<|header_end|>{message.content}<|eot|>"
        )
    elif isinstance(message, SystemMessage):
        message_text = f"<|header_start|>system<|header_end|>{message.content}<|eot|>"
    elif isinstance(message, ToolMessage):
        message_text = f"<|header_start|>ipython<|header_end|>{message.content}<|eom|>"
    else:
        raise ValueError(f"Got unknown type {message}")

    return message_text


def convert_messages_to_prompt_llama4(messages: List[BaseMessage]) -> str:
    """Convert a list of messages to a prompt for Llama 4."""

    return "\n".join(
        ["<|begin_of_text|>"]
        + [_convert_one_message_to_text_llama4(message) for message in messages]
        + ["<|header_start|>assistant<|header_end|>"]
    )


def _convert_one_message_to_text_anthropic(
    message: BaseMessage,
    human_prompt: str,
    ai_prompt: str,
) -> str:
    content = cast(str, message.content)
    if isinstance(message, ChatMessage):
        message_text = f"\n\n{message.role.capitalize()}: {content}"
    elif isinstance(message, HumanMessage):
        message_text = f"{human_prompt} {content}"
    elif isinstance(message, AIMessage):
        message_text = f"{ai_prompt} {content}"
    elif isinstance(message, SystemMessage):
        message_text = content
    else:
        raise ValueError(f"Got unknown type {message}")
    return message_text


def convert_messages_to_prompt_anthropic(
    messages: List[BaseMessage],
    *,
    human_prompt: str = "\n\nHuman:",
    ai_prompt: str = "\n\nAssistant:",
) -> str:
    """Format a list of messages into a full prompt for the Anthropic model
    Args:
        messages (List[BaseMessage]): List of BaseMessage to combine.
        human_prompt (str, optional): Human prompt tag. Defaults to "\n\nHuman:".
        ai_prompt (str, optional): AI prompt tag. Defaults to "\n\nAssistant:".
    Returns:
        str: Combined string with necessary human_prompt and ai_prompt tags.
    """
    if messages is None:
        return ""

    messages = messages.copy()  # don't mutate the original list
    if len(messages) > 0 and not isinstance(messages[-1], AIMessage):
        messages.append(AIMessage(content=""))

    text = "".join(
        _convert_one_message_to_text_anthropic(message, human_prompt, ai_prompt)
        for message in messages
    )

    # trim off the trailing ' ' that might come from the "Assistant: "
    return text.rstrip()


def _convert_one_message_to_text_mistral(message: BaseMessage) -> str:
    if isinstance(message, ChatMessage):
        message_text = f"\n\n{message.role.capitalize()}: {message.content}"
    elif isinstance(message, HumanMessage):
        message_text = f"[INST] {message.content} [/INST]"
    elif isinstance(message, AIMessage):
        message_text = f"{message.content}"
    elif isinstance(message, SystemMessage):
        message_text = f"<<SYS>> {message.content} <</SYS>>"
    else:
        raise ValueError(f"Got unknown type {message}")
    return message_text


def convert_messages_to_prompt_mistral(messages: List[BaseMessage]) -> str:
    """Convert a list of messages to a prompt for mistral."""
    return "\n".join(
        [_convert_one_message_to_text_mistral(message) for message in messages]
    )


def _convert_one_message_to_text_deepseek(message: BaseMessage) -> str:
    if isinstance(message, ChatMessage):
        message_text = f"<|{message.role}|>{message.content}"
    elif isinstance(message, HumanMessage):
        message_text = f"<|User|>{message.content}"
    elif isinstance(message, AIMessage):
        message_text = f"<|Assistant|>{message.content}"
    elif isinstance(message, SystemMessage):
        message_text = f"<|System|>{message.content}"
    else:
        raise ValueError(f"Got unknown type {message}")

    return message_text


def convert_messages_to_prompt_deepseek(messages: List[BaseMessage]) -> str:
    """Convert a list of messages to a prompt for DeepSeek-R1."""
    prompt = "\n<|begin_of_sentence|>"

    for message in messages:
        prompt += _convert_one_message_to_text_deepseek(message)

    prompt += "<|Assistant|>\n\n"

    return prompt


def _convert_one_message_to_text_writer(message: BaseMessage) -> str:
    if isinstance(message, ChatMessage):
        message_text = f"\n\n{message.role.capitalize()}: {message.content}"
    elif isinstance(message, HumanMessage):
        message_text = f"[INST] {message.content} [/INST]"
    elif isinstance(message, AIMessage):
        message_text = f"{message.content}"
    elif isinstance(message, SystemMessage):
        message_text = f"<<SYS>> {message.content} <</SYS>>"
    else:
        raise ValueError(f"Got unknown type {message}")
    return message_text


def convert_messages_to_prompt_writer(messages: List[BaseMessage]) -> str:
    """Convert a list of messages to a prompt for Writer."""

    return "\n".join(
        [_convert_one_message_to_text_writer(message) for message in messages]
    )


def _convert_one_message_to_text_openai(message: BaseMessage) -> str:
    if isinstance(message, SystemMessage):
        message_text = f"<|start|>system<|message|>{message.content}<|end|>"
    elif isinstance(message, ChatMessage):
        # developer role messages
        message_text = f"<|start|>{message.role}<|message|>{message.content}<|end|>"
    elif isinstance(message, HumanMessage):
        message_text = f"<|start|>user<|message|>{message.content}<|end|>"
    elif isinstance(message, AIMessage):
        message_text = (
            f"<|start|>assistant<|channel|>final<|message|>{message.content}<|end|>"
        )
    elif isinstance(message, ToolMessage):
        # TODO: Tool messages in the OpenAI format should use
        # "<|start|>{toolname} to=assistant<|message|>"
        # Need to extract the tool name from the ToolMessage content or tool_call_id
        # For now using generic "to=assistant" format as placeholder until we implement
        # tool calling
        # Will be resolved in follow-up PR with full tool support
        message_text = f"<|start|>to=assistant<|channel|>commentary<|message|>{message.content}<|end|>"  # noqa: E501
    else:
        raise ValueError(f"Got unknown type {message}")

    return message_text


def convert_messages_to_prompt_openai(messages: List[BaseMessage]) -> str:
    """Convert a list of messages to a Harmony format prompt for OpenAI API."""

    prompt = "\n"
    for message in messages:
        prompt += _convert_one_message_to_text_openai(message)

    prompt += "<|start|>assistant\n\n"

    return prompt


def _format_image(image_url: str) -> Dict:
    """Formats an image of format data:image/jpeg;base64,{b64_string}
    to a dict for anthropic api

    {
      "type": "base64",
      "media_type": "image/jpeg",
      "data": "/9j/4AAQSkZJRg...",
    }

    And throws an error if it's not a b64 image

    """
    regex = r"^data:(?P<media_type>image/.+);base64,(?P<data>.+)$"
    match = re.match(regex, image_url)
    if match is None:
        raise ValueError(
            "Anthropic only supports base64-encoded images currently."
            " Example: data:image/png;base64,'/9j/4AAQSk'..."
        )
    return {
        "type": "base64",
        "media_type": match.group("media_type"),
        "data": match.group("data"),
    }


def _format_data_content_block(block: dict) -> dict:
    """Format standard data content block to format expected by Converse API."""
    if block["type"] == "image":
        if "base64" in block or block.get("source_type") == "base64":
            if "mime_type" not in block:
                error_message = "mime_type key is required for base64 data."
                raise ValueError(error_message)
            formatted_block = {
                "type": "image",
                "source": {
                    "type": "base64",
                    "media_type": block["mime_type"],
                    "data": block.get("base64") or block.get("data", ""),
                },
            }
        else:
            error_message = "Image data only supported through in-line base64 format."
            raise ValueError(error_message)

    else:
        error_message = f"Blocks of type {block['type']} not supported."
        raise ValueError(error_message)

    return formatted_block


def _merge_messages(
    messages: Sequence[BaseMessage],
) -> List[Union[SystemMessage, AIMessage, HumanMessage]]:
    """Merge runs of human/tool messages into single human messages with content blocks."""  # noqa: E501
    merged: list = []
    for curr in messages:
        curr = curr.model_copy(deep=True)
        if isinstance(curr, ToolMessage):
            if isinstance(curr.content, list) and all(
                isinstance(block, dict) and block.get("type") == "tool_result"
                for block in curr.content
            ):
                curr = HumanMessage(curr.content)  # type: ignore[misc]
            else:
                curr = HumanMessage(  # type: ignore[misc]
                    [
                        {
                            "type": "tool_result",
                            "content": curr.content,
                            "tool_use_id": curr.tool_call_id,
                        }
                    ]
                )
        last = merged[-1] if merged else None
        if last is not None and any(
            all(isinstance(m, c) for m in (curr, last))
            for c in (SystemMessage, HumanMessage)
        ):
            if isinstance(last.content, str):
                new_content: List = [{"type": "text", "text": last.content}]
            else:
                new_content = last.content
            if isinstance(curr.content, str):
                new_content.append({"type": "text", "text": curr.content})
            else:
                new_content.extend(curr.content)
            last.content = new_content
        else:
            merged.append(curr)
    return merged


def _format_anthropic_messages(
    messages: List[BaseMessage],
) -> Tuple[Optional[Union[str, List[Dict[str, Any]]]], List[Dict[str, Any]]]:
    """Format messages for anthropic."""
    for idx, message in enumerate(messages):
        # Translate v1 content
        if (
            isinstance(message, AIMessage)
            and message.response_metadata.get("output_version") == "v1"
        ):
            messages[idx] = message.model_copy(
                update={
                    "content": _convert_from_v1_to_anthropic(
                        cast(list[types.ContentBlock], message.content),
                        message.response_metadata.get("model_provider"),
                    )
                }
            )
    system: Optional[Union[str, List[Dict[str, Any]]]] = None
    formatted_messages: List[Dict[str, Any]] = []

    trimmed_messages = trim_message_whitespace(messages)
    merged_messages = _merge_messages(trimmed_messages)
    for i, message in enumerate(merged_messages):
        if message.type == "system":
            if system is not None:
                raise ValueError("Received multiple non-consecutive system messages.")
            elif isinstance(message.content, str):
                system = message.content
            elif isinstance(message.content, list):
                system_blocks = []
                for item in message.content:
                    if isinstance(item, str):
                        system_blocks.append({"type": "text", "text": item})
                    elif isinstance(item, dict):
                        if item.get("type") != "text":
                            raise ValueError(
                                "System message content item must be type 'text'"
                            )
                        if "text" not in item:
                            raise ValueError(
                                "System message content item must have a 'text' key"
                            )
                        content_item = {"type": "text", "text": item["text"]}
                        if item.get("cache_control"):
                            content_item["cache_control"] = {"type": "ephemeral"}
                        system_blocks.append(content_item)
                    else:
                        raise ValueError(
                            "System message content list must be a string or dict, "
                            f"instead was: {type(item)}"
                        )
                system = system_blocks
            else:
                raise ValueError(
                    "System message content must be a string or list, "
                    f"instead was: {type(message.content)}"
                )

            continue

        role = _message_type_lookups[message.type]
        final_content: Union[str, List[Dict[str, Any]]]

        if not isinstance(message.content, str):
            # parse as dict
            assert isinstance(message.content, list), (
                "Anthropic message content must be str or list of dicts"
            )

            # populate content
            thinking_blocks: List[Dict[str, Any]] = []
            native_blocks: List[Dict[str, Any]] = []
            tool_blocks: List[Dict[str, Any]] = []

            # First collect all blocks by type
            for item in message.content:
                if isinstance(item, str):
                    native_blocks.append({"type": "text", "text": item})
                elif isinstance(item, dict):
                    if "type" not in item:
                        raise ValueError("Dict content item must have a type key")
                    elif is_data_content_block(item):
                        native_blocks.append(_format_data_content_block(item))
                    elif item["type"] == "image_url":
                        # convert format
                        source = _format_image(item["image_url"]["url"])
                        native_blocks.append({"type": "image", "source": source})  # type: ignore
                    elif item["type"] == "image":
                        native_blocks.append(item)
                    elif item["type"] == "tool_result":
                        # Process content within tool_result
                        content_item = item["content"]
                        if isinstance(content_item, list):
                            # Handle list content inside tool_result
                            processed_list = []
                            for list_item in content_item:
                                if (
                                    isinstance(list_item, dict)
                                    and list_item.get("type") == "image_url"
                                ):
                                    # Process image in list
                                    source = _format_image(
                                        list_item["image_url"]["url"]
                                    )
                                    processed_list.append(
                                        {"type": "image", "source": source}
                                    )
                                elif (
                                    isinstance(list_item, dict)
                                    and list_item.get("type") == "text"
                                ):
                                    # Strip extra fields that are not accepted
                                    # by the Bedrock API.
                                    formatted_item: Dict[str, Any] = {
                                        k: v
                                        for k, v in list_item.items()
                                        if k
                                        in (
                                            "type",
                                            "text",
                                            "cache_control",
                                            "citations",
                                        )
                                    }
                                    processed_list.append(formatted_item)
                                else:
                                    # Keep other items as is
                                    processed_list.append(list_item)
                            # Add processed list to tool_result
                            tool_blocks.append(
                                {
                                    "type": "tool_result",
                                    "tool_use_id": item.get("tool_use_id"),
                                    "content": processed_list,
                                }
                            )
                        else:
                            # For other content types, keep as is
                            tool_blocks.append(item)
                    elif item["type"] == "tool_use":
                        # If a tool_call with the same id as a tool_use content block
                        # exists, the tool_call is preferred.
                        if isinstance(message, AIMessage) and item["id"] in [
                            tc["id"] for tc in message.tool_calls
                        ]:
                            overlapping = [
                                tc
                                for tc in message.tool_calls
                                if tc["id"] == item["id"]
                            ]
                            tool_blocks.extend(
                                cast(
                                    List[Dict[str, Any]],
                                    _lc_tool_calls_to_anthropic_tool_use_blocks(
                                        overlapping
                                    ),
                                )
                            )
                        else:
                            item.pop("text", None)
                            tool_blocks.append(item)
                    elif item["type"] in ["thinking", "redacted_thinking"]:
                        # Store thinking blocks separately
                        thinking_blocks.append(
                            {k: v for k, v in item.items() if k != "index"}
                        )
                    elif item["type"] == "text":
                        text = item.get("text", "")
                        # Only add non-empty strings for now as empty ones are not
                        # accepted.
                        # https://github.com/anthropics/anthropic-sdk-python/issues/461

                        if text.strip():
                            content_item = {"type": "text", "text": text}
                            if item.get("cache_control"):
                                content_item["cache_control"] = {"type": "ephemeral"}
                            if item.get("citations"):
                                content_item["citations"] = item["citations"]
                            native_blocks.append(content_item)
                    else:
                        tool_blocks.append(item)
                else:
                    raise ValueError(
                        f"Content items must be str or dict, instead was: {type(item)}"
                    )

            # Add tool calls if present (for AIMessage)
            if isinstance(message, AIMessage) and message.tool_calls:
                # Track which tool call IDs we've already processed
                used_tool_call_ids = {
                    block["id"]
                    for block in tool_blocks
                    if block.get("type") == "tool_use"
                }
                # Only add tool calls that haven't been processed yet
                new_tool_calls = [
                    tc
                    for tc in message.tool_calls
                    if tc["id"] not in used_tool_call_ids
                ]
                if new_tool_calls:
                    tool_blocks.extend(
                        cast(
                            List[Dict[str, Any]],
                            _lc_tool_calls_to_anthropic_tool_use_blocks(new_tool_calls),
                        )
                    )

            # For assistant messages, when thinking blocks exist, ensure they come first
            if role == "assistant":
                final_content = native_blocks + tool_blocks
                if thinking_blocks:
                    final_content = thinking_blocks + final_content
            elif role == "user" and tool_blocks and native_blocks:
                final_content = (
                    tool_blocks + native_blocks
                )  # tool result must precede text
                if thinking_blocks:
                    final_content = thinking_blocks + final_content
            else:
                # combine all blocks in standard order
                final_content = native_blocks + tool_blocks
                # Only include thinking blocks if they exist
                if thinking_blocks:
                    final_content = thinking_blocks + final_content

        elif isinstance(message, AIMessage):
            # For string content, create appropriate structure
            content_list = []

            # Add thinking blocks from additional_kwargs if present
            if message.additional_kwargs and "thinking" in message.additional_kwargs:
                thinking_data = message.additional_kwargs["thinking"]
                if thinking_data and isinstance(thinking_data, dict):
                    if "text" in thinking_data and "signature" in thinking_data:
                        content_list.append(
                            {
                                "type": "thinking",
                                "thinking": thinking_data["text"],
                                "signature": thinking_data["signature"],
                            }
                        )

            # Add base content as text block
            if message.content:
                content_list.append({"type": "text", "text": message.content})

            # Add tool calls if present
            if message.tool_calls:
                content_list.extend(
                    cast(
                        List[Dict[str, Any]],
                        _lc_tool_calls_to_anthropic_tool_use_blocks(message.tool_calls),
                    )
                )

            # For assistant messages with thinking blocks, ensure they come first
            if role == "assistant" and any(
                block.get("type") in ["thinking", "redacted_thinking"]
                for block in content_list
                if isinstance(block, dict)
            ):
                # Separate thinking blocks and non-thinking blocks
                thinking_blocks = [
                    block
                    for block in content_list
                    if isinstance(block, dict)
                    and block.get("type") in ["thinking", "redacted_thinking"]
                ]
                other_blocks = [
                    block
                    for block in content_list
                    if not (
                        isinstance(block, dict)
                        and block.get("type") in ["thinking", "redacted_thinking"]
                    )
                ]
                # Combine with thinking first
                final_content = thinking_blocks + other_blocks
            else:
                # No thinking blocks or not an assistant message
                final_content = content_list
        else:
            # Simple string content
            final_content = message.content

        # AWS Bedrock requires content arrays to have at least 1 item
        if isinstance(final_content, list) and len(final_content) == 0:
            final_content = [{"type": "text", "text": "."}]

        formatted_messages.append({"role": role, "content": final_content})
    return system, formatted_messages


class ChatPromptAdapter:
    """Adapter class to prepare the inputs from Langchain to prompt format that Chat
    model expects.

    """

    @classmethod
    def convert_messages_to_prompt(
        cls, provider: str, messages: List[BaseMessage], model: str
    ) -> str:
        if provider == "anthropic":
            prompt = convert_messages_to_prompt_anthropic(messages=messages)
        elif provider == "deepseek":
            prompt = convert_messages_to_prompt_deepseek(messages=messages)
        elif provider == "meta":
            if "llama4" in model:
                prompt = convert_messages_to_prompt_llama4(messages=messages)
            elif "llama3" in model:
                prompt = convert_messages_to_prompt_llama3(messages=messages)
            else:
                prompt = convert_messages_to_prompt_llama(messages=messages)
        elif provider == "mistral":
            prompt = convert_messages_to_prompt_mistral(messages=messages)
        elif provider == "amazon":
            prompt = convert_messages_to_prompt_anthropic(
                messages=messages,
                human_prompt="\n\nUser:",
                ai_prompt="\n\nBot:",
            )
        elif provider == "writer":
            prompt = convert_messages_to_prompt_writer(messages=messages)
        elif provider == "openai":
            prompt = convert_messages_to_prompt_openai(messages=messages)
        else:
            raise NotImplementedError(
                f"Provider {provider} model does not support chat."
            )
        return prompt

    @classmethod
    def format_messages(
        cls, provider: str, messages: List[BaseMessage]
    ) -> Union[
        Tuple[Optional[Union[str, List[Dict[str, Any]]]], List[Dict[str, Any]]],
        List[Dict[str, Any]],
    ]:
        if provider == "anthropic":
            return _format_anthropic_messages(messages)
        elif provider in ("openai", "qwen"):
            return cast(List[Dict[str, Any]], convert_to_openai_messages(messages))
        raise NotImplementedError(
            f"Provider {provider} not supported for format_messages"
        )


_message_type_lookups = {
    "human": "user",
    "ai": "assistant",
    "AIMessageChunk": "assistant",
    "HumanMessageChunk": "user",
}


def _apply_cache_control_to_messages(
    cache_control: Optional[Dict[str, Any]],
    formatted_messages: Optional[List[Dict[str, Any]]],
) -> None:
    """Apply cache_control to the last content block of the last message."""
    if not cache_control or not formatted_messages:
        return
    for fmt_msg in reversed(formatted_messages):
        content = fmt_msg.get("content")
        if isinstance(content, list) and content:
            for block in reversed(content):
                if isinstance(block, dict):
                    block["cache_control"] = cache_control
                    break
            break
        elif isinstance(content, str):
            fmt_msg["content"] = [
                {
                    "type": "text",
                    "text": content,
                    "cache_control": cache_control,
                }
            ]
            break


class ChatBedrock(BaseChatModel, BedrockBase):
    """A chat model that uses the Bedrock API."""

    system_prompt_with_tools: str = ""
    beta_use_converse_api: bool = False
    """Use the new Bedrock `converse` API which provides a standardized interface to
    all Bedrock models. Support still in beta. See ChatBedrockConverse docs for more."""

    stop_sequences: Optional[List[str]] = Field(default=None, alias="stop")
    """Stop sequence inference parameter from new Bedrock `converse` API providing
    a sequence of characters that causes a model to stop generating a response. See
    https://docs.aws.amazon.com/bedrock/latest/APIReference/API_agent_InferenceConfiguration.html
    for more.

    """

    @property
    def _llm_type(self) -> str:
        """Return type of chat model."""
        return "amazon_bedrock_chat"

    @classmethod
    def is_lc_serializable(cls) -> bool:
        """Return whether this model can be serialized by Langchain."""
        return True

    @classmethod
    def get_lc_namespace(cls) -> List[str]:
        """Get the namespace of the langchain object.

        Returns:
            `["langchain", "chat_models", "bedrock"]`
        """
        return ["langchain", "chat_models", "bedrock"]

    @model_validator(mode="before")
    @classmethod
    def set_beta_use_converse_api(cls, values: Dict) -> Any:
        model_id = values.get("model_id", values.get("model"))
        base_model_id = values.get("base_model_id", values.get("base_model", ""))

        if not model_id or "beta_use_converse_api" in values:
            return values

        nova_id = "amazon.nova"
        values["beta_use_converse_api"] = False

        if nova_id in model_id or nova_id in base_model_id:
            values["beta_use_converse_api"] = True
        elif not base_model_id and "application-inference-profile" in model_id:
            bedrock_client = values.get("bedrock_client")
            if not bedrock_client:
                bedrock_client = create_aws_client(
                    region_name=values.get("region_name"),
                    credentials_profile_name=values.get("credentials_profile_name"),
                    aws_access_key_id=values.get("aws_access_key_id"),
                    aws_secret_access_key=values.get("aws_secret_access_key"),
                    aws_session_token=values.get("aws_session_token"),
                    endpoint_url=values.get("endpoint_url"),
                    config=values.get("config"),
                    service_name="bedrock",
                )
            response = bedrock_client.get_inference_profile(
                inferenceProfileIdentifier=model_id
            )
            if "models" in response and len(response["models"]) > 0:
                model_arn = response["models"][0]["modelArn"]
                resolved_base_model = model_arn.split("/")[-1]
                values["beta_use_converse_api"] = "nova" in resolved_base_model
        return values

    @model_validator(mode="before")
    @classmethod
    def build_extra(cls, values: dict[str, Any]) -> Any:
        """Build extra kwargs from additional params that were passed in."""
        all_required_field_names = get_pydantic_field_names(cls)

        # For backwards compatibility, we don't transfer known parameters out of
        # model_kwargs
        model_kwargs = values.pop("model_kwargs", {})
        values = _build_model_kwargs(values, all_required_field_names)
        if model_kwargs or values.get("model_kwargs", {}):
            values["model_kwargs"] = {
                **values.get("model_kwargs", {}),
                **model_kwargs,
            }
        return values

    @model_validator(mode="after")
    def _set_model_profile(self) -> Self:
        """Set model profile if not overridden."""
        if self.profile is None:
            model_id = re.sub(r"^[A-Za-z]{2}\.", "", self.model_id)
            self.profile = _get_default_model_profile(model_id)
        return self

    @property
    def lc_attributes(self) -> Dict[str, Any]:
        attributes: Dict[str, Any] = {}

        if self.region_name:
            attributes["region_name"] = self.region_name

        return attributes

    model_config = ConfigDict(
        extra="forbid",
        populate_by_name=True,
    )

    def _get_ls_params(
        self, stop: Optional[List[str]] = None, **kwargs: Any
    ) -> LangSmithParams:
        """Get standard params for tracing."""
        params = self._get_invocation_params(stop=stop, **kwargs)
        ls_params = LangSmithParams(
            ls_provider="amazon_bedrock",
            ls_model_name=self.model_id,
            ls_model_type="chat",
        )
        if ls_temperature := params.get("temperature", self.temperature):
            ls_params["ls_temperature"] = ls_temperature
        if ls_max_tokens := params.get("max_tokens", self.max_tokens):
            ls_params["ls_max_tokens"] = ls_max_tokens
        if ls_stop := stop or params.get("stop", None):
            ls_params["ls_stop"] = ls_stop
        return ls_params

    def _stream(
        self,
        messages: List[BaseMessage],
        stop: Optional[List[str]] = None,
        run_manager: Optional[CallbackManagerForLLMRun] = None,
        **kwargs: Any,
    ) -> Iterator[ChatGenerationChunk]:
        if self.beta_use_converse_api:
            yield from self._as_converse._stream(
                messages, stop=stop, run_manager=run_manager, **kwargs
            )
            return
        provider = self._get_provider()
        prompt: Optional[str] = None
        system: Optional[Union[str, List[Dict[str, Any]]]] = None
        formatted_messages: Optional[List[Dict[str, Any]]] = None

        if provider == "anthropic":
            result = ChatPromptAdapter.format_messages(provider, messages)
            assert isinstance(result, tuple)
            system_raw, formatted_messages = (
                result[0],
                cast(List[Dict[str, Any]], result[1]),
            )
            # Preserve system prompt format (str or list) for cache_control support
            if system_raw:
                system = system_raw

            if self.system_prompt_with_tools:
                if system is None:
                    system = self.system_prompt_with_tools
                elif isinstance(system, str):
                    system = f"{self.system_prompt_with_tools}\n{system}"
                else:
                    # Prepend tools as a content block to preserve cache_control
                    system = [
                        {"type": "text", "text": self.system_prompt_with_tools}
                    ] + list(system)

            # Apply cache_control to last message if provided via kwargs
            _apply_cache_control_to_messages(
                kwargs.pop("cache_control", None), formatted_messages
            )
        elif provider in ("openai", "qwen"):
            formatted_messages = cast(
                List[Dict[str, Any]],
                ChatPromptAdapter.format_messages(provider, messages),
            )
        else:
            prompt = ChatPromptAdapter.convert_messages_to_prompt(
                provider=provider, messages=messages, model=self._get_base_model()
            )

        added_model_name = False
        # Track guardrails trace information for callback handling
        guardrails_trace_info = None

        for chunk in self._prepare_input_and_invoke_stream(
            prompt=prompt,
            system=system,
            messages=formatted_messages,
            stop=stop,
            run_manager=run_manager,
            **kwargs,
        ):
            if isinstance(chunk, AIMessageChunk):
                chunk.response_metadata["model_provider"] = "bedrock"
                generation_chunk = ChatGenerationChunk(message=chunk)
                if run_manager:
                    run_manager.on_llm_new_token(
                        generation_chunk.text, chunk=generation_chunk
                    )
                yield generation_chunk
            else:
                delta = chunk.text
                response_metadata = None
                if generation_info := chunk.generation_info:
                    # Check for guardrail intervention in the streaming chunk
                    services_trace = self._get_bedrock_services_signal(generation_info)
                    if services_trace.get("signal") and run_manager:
                        # Store trace info for potential callback
                        guardrails_trace_info = services_trace

                    usage_metadata = generation_info.pop("usage_metadata", None)
                    response_metadata = generation_info
                    if not added_model_name:
                        response_metadata["model_name"] = self.model_id
                        added_model_name = True
                else:
                    usage_metadata = None
                generation_chunk = ChatGenerationChunk(
                    message=AIMessageChunk(
                        content=delta,
                        response_metadata=response_metadata,
                        usage_metadata=usage_metadata,
                    )
                    if response_metadata is not None
                    else AIMessageChunk(content=delta)
                )
                generation_chunk.message.response_metadata["model_provider"] = "bedrock"
                if run_manager:
                    run_manager.on_llm_new_token(
                        generation_chunk.text, chunk=generation_chunk
                    )
                yield generation_chunk

        # If guardrails intervened during streaming, notify the callback handler
        if guardrails_trace_info and run_manager:
            run_manager.on_llm_error(
                Exception(
                    f"Error raised by bedrock service: "
                    f"{guardrails_trace_info.get('reason')}"
                ),
                **guardrails_trace_info,
            )

    def _generate(
        self,
        messages: List[BaseMessage],
        stop: Optional[List[str]] = None,
        run_manager: Optional[CallbackManagerForLLMRun] = None,
        **kwargs: Any,
    ) -> ChatResult:
        if self.beta_use_converse_api:
            if not self.streaming:
                return self._as_converse._generate(
                    messages, stop=stop, run_manager=run_manager, **kwargs
                )
            else:
                stream_iter = self._as_converse._stream(
                    messages, stop=stop, run_manager=run_manager, **kwargs
                )
                return generate_from_stream(stream_iter)
        completion = ""
        llm_output: Dict[str, Any] = {}
        tool_calls: List[ToolCall] = []
        citations_enabled: Optional[bool] = None
        provider_stop_reason_code = self.provider_stop_reason_key_map.get(
            self._get_provider(), "stop_reason"
        )
        provider = self._get_provider()
        if self.streaming:
            if provider == "anthropic":
                stream_iter = self._stream(messages, stop, run_manager, **kwargs)
                return generate_from_stream(stream_iter)

            response_metadata: List[Dict[str, Any]] = []
            for chunk in self._stream(messages, stop, run_manager, **kwargs):
                completion += chunk.text
                response_metadata.append(chunk.message.response_metadata)
                if "tool_calls" in chunk.message.additional_kwargs.keys():
                    tool_calls = chunk.message.additional_kwargs["tool_calls"]
            llm_output = _combine_generation_info_for_llm_result(
                response_metadata, provider_stop_reason_code
            )
        else:
            prompt: Optional[str] = None
            system: Optional[Union[str, List[Dict[str, Any]]]] = None
            formatted_messages: Optional[List[Dict[str, Any]]] = None
            params: Dict[str, Any] = {**kwargs}

            if provider == "anthropic":
                result = ChatPromptAdapter.format_messages(provider, messages)
                assert isinstance(result, tuple)
                system_raw, formatted_messages = (
                    result[0],
                    cast(List[Dict[str, Any]], result[1]),
                )
                # Preserve system prompt format (str or list) for cache_control support
                if system_raw:
                    system = system_raw

                if self.system_prompt_with_tools:
                    if system is None:
                        system = self.system_prompt_with_tools
                    elif isinstance(system, str):
                        system = f"{self.system_prompt_with_tools}\n{system}"
                    else:
                        # Prepend tools as a content block to preserve cache_control
                        system = [
                            {"type": "text", "text": self.system_prompt_with_tools}
                        ] + list(system)
                citations_enabled = _citations_enabled(formatted_messages)

                # Apply cache_control to last message if provided via kwargs
                _apply_cache_control_to_messages(
                    params.pop("cache_control", None), formatted_messages
                )

            elif provider in ("openai", "qwen"):
                formatted_messages = cast(
                    List[Dict[str, Any]],
                    ChatPromptAdapter.format_messages(provider, messages),
                )
            else:
                prompt = ChatPromptAdapter.convert_messages_to_prompt(
                    provider=provider, messages=messages, model=self._get_base_model()
                )

            if stop:
                params["stop_sequences"] = stop

            completion, tool_calls, llm_output, body = self._prepare_input_and_invoke(
                prompt=prompt,
                stop=stop,
                run_manager=run_manager,
                system=system,
                messages=formatted_messages,
                **params,
            )
        # usage metadata
        if usage := llm_output.get("usage"):
            input_tokens = usage.get("prompt_tokens", 0)
            output_tokens = usage.get("completion_tokens", 0)
            cache_read_input_tokens = usage.get("cache_read_input_tokens", 0)
            cache_write_input_tokens = usage.get("cache_write_input_tokens", 0)
            usage_metadata = UsageMetadata(
                input_tokens=input_tokens,
                output_tokens=output_tokens,
                input_token_details={
                    "cache_read": cache_read_input_tokens,
                    "cache_creation": cache_write_input_tokens,
                },
                total_tokens=usage.get("total_tokens", input_tokens + output_tokens),
            )
        else:
            usage_metadata = None
        logger.debug(f"The message received from Bedrock: {completion}")
        llm_output["model_id"] = self.model_id  # backward-compatibility

        # Use raw response content in some cases, so that thinking and citations
        # are properly stored in content array
        content = completion
        if (response_content := body.get("content")) and (
            (_ := llm_output.pop("thinking", None)) or citations_enabled
        ):
            content = response_content

        msg = AIMessage(
            content=content,
            additional_kwargs=llm_output,
            tool_calls=cast(List[ToolCall], tool_calls),
            usage_metadata=usage_metadata,
            response_metadata={
                "model_provider": "bedrock",
                "model_name": self.model_id,
            },
        )
        return ChatResult(
            generations=[
                ChatGeneration(
                    message=msg,
                )
            ],
            llm_output=llm_output,
        )

    def _combine_llm_outputs(self, llm_outputs: List[Optional[dict]]) -> dict:
        final_usage: Dict[str, int] = defaultdict(int)
        final_output = {}
        for output in llm_outputs:
            output = output or {}
            usage = output.get("usage", {})
            for token_type, token_count in usage.items():
                final_usage[token_type] += token_count
            final_output.update(output)
        final_output["usage"] = final_usage
        return final_output

    def get_num_tokens_from_messages(
        self, messages: list[BaseMessage], tools: Optional[Sequence] = None
    ) -> int:
        model_id = self._get_base_model()
        if self._model_is_anthropic and count_tokens_api_supported_for_model(model_id):
            system, formatted_messages = ChatPromptAdapter.format_messages(
                "anthropic", messages
            )
            input_to_count_tmpl = {
                "anthropic_version": "bedrock-2023-05-31",
                "max_tokens": self.max_tokens if self.max_tokens else 8192,
                "messages": formatted_messages,
            }
            if system:
                input_to_count_tmpl["system"] = system
            input_to_count = json.dumps(input_to_count_tmpl)

            response = self.client.count_tokens(
                modelId=model_id, input={"invokeModel": {"body": input_to_count}}
            )
            return response["inputTokens"]

        return super().get_num_tokens_from_messages(messages, tools)

    def get_num_tokens(self, text: str) -> int:
        if (
            self._model_is_anthropic
            and not self.custom_get_token_ids
            and anthropic_tokens_supported()
        ):
            return get_num_tokens_anthropic(text)
        return super().get_num_tokens(text)

    def get_token_ids(self, text: str) -> List[int]:
        if self._model_is_anthropic and not self.custom_get_token_ids:
            if anthropic_tokens_supported():
                return get_token_ids_anthropic(text)
            else:
                warnings.warn(
                    "Falling back to default token method due to missing or "
                    "incompatible `anthropic` installation "
                    "(needs <=0.38.0).\n\nIf using `anthropic>0.38.0`, "
                    "it is recommended to provide the model class with a "
                    "custom_get_token_ids method implementing a more accurate "
                    "tokenizer for Anthropic. For get_num_tokens, as another "
                    "alternative, you can implement your own token counter method "
                    "using the ChatAnthropic or AnthropicLLM classes."
                )
        return super().get_token_ids(text)

    def set_system_prompt_with_tools(self, xml_tools_system_prompt: str) -> None:
        """Workaround to bind. Sets the system prompt with tools"""
        self.system_prompt_with_tools = xml_tools_system_prompt

    def bind_tools(
        self,
        tools: Sequence[Union[Dict[str, Any], TypeBaseModel, Callable, BaseTool]],
        *,
        tool_choice: Optional[Union[dict, str, Literal["auto", "none"], bool]] = None,
        strict: Optional[bool] = None,
        **kwargs: Any,
    ) -> Runnable[LanguageModelInput, AIMessage]:
        """Bind tool-like objects to this chat model.

        Assumes model has a tool calling API.

        Args:
            tools: A list of tool definitions to bind to this chat model.
                Can be  a dictionary, pydantic model, callable, or BaseTool. Pydantic
                models, callables, and BaseTools will be automatically converted to
                their schema dictionary representation.
            tool_choice: Which tool to require the model to call.
                Must be the name of the single provided function or
                "auto" to automatically determine which function to call
                (if any), or a dict of the form:
                {"type": "function", "function": {"name": <<tool_name>>}}.
            strict: If True, enables strict mode for tool definitions.
                Only supported when Converse API passthrough is enabled.
                (beta_use_converse_api=True).
            **kwargs: Any additional parameters to pass to the
                [Runnable][langchain_core.runnables.Runnable] constructor.

        """
        if self.beta_use_converse_api:
            if isinstance(tool_choice, bool):
                tool_choice = "any" if tool_choice else None
            return self._as_converse.bind_tools(
                tools, tool_choice=tool_choice, strict=strict, **kwargs
            )
        if strict is not None:
            warnings.warn(
                "The 'strict' parameter is only supported when using the Converse "
                "API (beta_use_converse_api=True). It will be ignored.",
                UserWarning,
                stacklevel=2,
            )
        if self._get_provider() == "anthropic":
            formatted_tools = [convert_to_anthropic_tool(tool) for tool in tools]

            base_model = self._get_base_model()
            if any(
                x in base_model
                for x in (
                    "claude-3-7-",
                    "claude-opus-4-",
                    "claude-sonnet-4-",
                    "claude-haiku-4-",
                )
            ) and thinking_in_params(self.model_kwargs or {}):
                forced = False
                if isinstance(tool_choice, bool):
                    forced = bool(tool_choice)
                elif isinstance(tool_choice, str):
                    # "any" or specific tool name forces tool use; "auto"/"none" do not
                    if tool_choice == "any":
                        forced = True
                    elif tool_choice not in ("auto", "none"):
                        # Treat as specific tool name
                        forced = True
                elif isinstance(tool_choice, dict) and tool_choice is not None:
                    tc_type = tool_choice.get("type")
                    # Bedrock types: "auto", "any", "tool" (function)
                    if tc_type in ("any", "tool", "function"):
                        forced = True
                if forced:
                    raise ValueError(
                        "Anthropic Claude (3.7/4/4.1/4.5) with thinking enabled does "
                        "not support forced tool use. Remove forced tool_choice (e.g. "
                        "'any' or a specific tool), or set tool_choice='auto', or "
                        "disable thinking."
                    )

            # true if the model is a claude 3 model
            if "claude-" in self._get_base_model():
                if not tool_choice:
                    pass
                elif isinstance(tool_choice, dict):
                    kwargs["tool_choice"] = tool_choice
                elif isinstance(tool_choice, str) and tool_choice in ("any", "auto"):
                    kwargs["tool_choice"] = {"type": tool_choice}
                elif isinstance(tool_choice, str):
                    kwargs["tool_choice"] = {"type": "tool", "name": tool_choice}
                else:
                    raise ValueError(
                        f"Unrecognized 'tool_choice' type {tool_choice=}."
                        f"Expected dict, str, or None."
                    )
                return self.bind(tools=formatted_tools, **kwargs)
            else:
                # add tools to the system prompt, the old way
                system_formatted_tools = get_system_message(
                    cast(List[AnthropicTool], formatted_tools)
                )
                self.set_system_prompt_with_tools(system_formatted_tools)
        return self

    def with_structured_output(
        self,
        schema: Union[Dict[str, Any], TypeBaseModel, Type],
        *,
        include_raw: bool = False,
        strict: Optional[bool] = None,
        **kwargs: Any,
    ) -> Runnable[LanguageModelInput, Union[Dict, BaseModel]]:
        """Model wrapper that returns outputs formatted to match the given schema.

        Args:
            schema: The output schema as a dict or a Pydantic class. If a Pydantic class
                then the model output will be an object of that class. If a dict then
                the model output will be a dict. With a Pydantic class the returned
                attributes will be validated, whereas with a dict they will not be.
            include_raw:
                If `False` then only the parsed structured output is returned.

                If an error occurs during model output parsing it will be raised.

                If `True` then both the raw model response (a `BaseMessage`) and the
                parsed model response will be returned.

                If an error occurs during output parsing it will be caught and returned
                as well.

                The final output is always a `dict` with keys `'raw'`, `'parsed'`, and
                `'parsing_error'`.
            strict: If True, enables strict mode for tool definitions.
                Only supported when Converse API passthrough is enabled
                (beta_use_converse_api=True).

        Returns:
            A Runnable that takes any ChatModel input. The output type depends on
            include_raw and schema.

            If include_raw is True then output is a dict with keys:
                raw: BaseMessage,
                parsed: Optional[_DictOrPydantic],
                parsing_error: Optional[BaseException],

            If include_raw is False and schema is a Dict then the runnable outputs a Dict.
            If include_raw is False and schema is a Type[BaseModel] then the runnable
            outputs a BaseModel.

        Example: Pydantic schema (include_raw=False):
            ```python
            from langchain_aws.chat_models.bedrock import ChatBedrock
            from pydantic import BaseModel

            class AnswerWithJustification(BaseModel):
                '''An answer to the user question along with justification for the answer.'''
                answer: str
                justification: str

            llm = ChatBedrock(
                model_id="anthropic.claude-3-sonnet-20240229-v1:0",
                model_kwargs={"temperature": 0.001},
            )  # type: ignore[call-arg]
            structured_model = model.with_structured_output(AnswerWithJustification)

            structured_model.invoke("What weighs more a pound of bricks or a pound of feathers")

            # -> AnswerWithJustification(
            #     answer='They weigh the same',
            #     justification='Both a pound of bricks and a pound of feathers weigh one pound. The weight is the same, but the volume or density of the objects may differ.'
            # )
            ```

        Example:  Pydantic schema (include_raw=True):
            ```python
            from langchain_aws.chat_models.bedrock import ChatBedrock
            from pydantic import BaseModel

            class AnswerWithJustification(BaseModel):
                '''An answer to the user question along with justification for the answer.'''
                answer: str
                justification: str

            model = ChatBedrock(
                model_id="anthropic.claude-3-sonnet-20240229-v1:0",
                model_kwargs={"temperature": 0.001},
            )  # type: ignore[call-arg]
            structured_model = model.with_structured_output(AnswerWithJustification, include_raw=True)

            structured_model.invoke("What weighs more a pound of bricks or a pound of feathers")
            # -> {
            #     'raw': AIMessage(content='', additional_kwargs={'tool_calls': [{'id': 'call_Ao02pnFYXD6GN1yzc0uXPsvF', 'function': {'arguments': '{"answer":"They weigh the same.","justification":"Both a pound of bricks and a pound of feathers weigh one pound. The weight is the same, but the volume or density of the objects may differ."}', 'name': 'AnswerWithJustification'}, 'type': 'function'}]}),
            #     'parsed': AnswerWithJustification(answer='They weigh the same.', justification='Both a pound of bricks and a pound of feathers weigh one pound. The weight is the same, but the volume or density of the objects may differ.'),
            #     'parsing_error': None
            # }
            ```

        Example: Dict schema (include_raw=False):
            ```python
            from langchain_aws.chat_models.bedrock import ChatBedrock

            schema = {
                "name": "AnswerWithJustification",
                "description": "An answer to the user question along with justification for the answer.",
                "input_schema": {
                    "type": "object",
                    "properties": {
                        "answer": {"type": "string"},
                        "justification": {"type": "string"},
                    },
                    "required": ["answer", "justification"]
                }
            }
            model = ChatBedrock(
                model_id="anthropic.claude-3-sonnet-20240229-v1:0",
                model_kwargs={"temperature": 0.001},
            )  # type: ignore[call-arg]
            structured_model = model.with_structured_output(schema)

            structured_model.invoke("What weighs more a pound of bricks or a pound of feathers")
            # -> {
            #     'answer': 'They weigh the same',
            #     'justification': 'Both a pound of bricks and a pound of feathers weigh one pound. The weight is the same, but the volume and density of the two substances differ.'
            # }
            ```

        """  # noqa: E501
        if self.beta_use_converse_api:
            return self._as_converse.with_structured_output(
                schema, include_raw=include_raw, strict=strict, **kwargs
            )
        if "claude-" not in self._get_base_model():
            raise ValueError(
                f"Structured output is not supported for model {self._get_base_model()}"
            )

        tool_name = convert_to_anthropic_tool(schema)["name"]

        base_model = self._get_base_model()
        has_thinking = any(
            x in base_model
            for x in (
                "claude-3-7-",
                "claude-opus-4-",
                "claude-sonnet-4-",
                "claude-haiku-4-",
            )
        ) and thinking_in_params(self.model_kwargs or {})

        if has_thinking:
            llm = self.bind_tools(
                [schema],
                strict=strict,
                ls_structured_output_format={
                    "kwargs": {"method": "function_calling"},
                    "schema": convert_to_openai_tool(schema),
                },
            )
        else:
            llm = self.bind_tools(
                [schema],
                tool_choice=tool_name,
                strict=strict,
                ls_structured_output_format={
                    "kwargs": {"method": "function_calling"},
                    "schema": convert_to_openai_tool(schema),
                },
            )
        if isinstance(schema, type) and is_basemodel_subclass(schema):
            if self.streaming:
                output_parser: OutputParserLike = PydanticToolsParser(
                    first_tool_only=True, tools=[schema]
                )
            else:
                output_parser = ToolsOutputParser(
                    first_tool_only=True, pydantic_schemas=[schema]
                )
        else:
            if self.streaming:
                output_parser = JsonOutputKeyToolsParser(
                    first_tool_only=True, key_name=tool_name
                )
            else:
                output_parser = ToolsOutputParser(first_tool_only=True, args_only=True)

        if has_thinking:

            def _raise_if_no_tool_calls(message: AIMessage) -> AIMessage:
                if not message.tool_calls:
                    raise OutputParserException(
                        "Model with thinking enabled was expected to call tools "
                        "but did not. This can happen when tool_choice is downgraded "
                        "from forced to 'auto' due to thinking mode limitations."
                    )
                return message

            llm = llm | _raise_if_no_tool_calls

        if include_raw:
            parser_assign = RunnablePassthrough.assign(
                parsed=itemgetter("raw") | output_parser, parsing_error=lambda _: None
            )
            parser_none = RunnablePassthrough.assign(parsed=lambda _: None)
            parser_with_fallback = parser_assign.with_fallbacks(
                [parser_none], exception_key="parsing_error"
            )
            return RunnableMap(raw=llm) | parser_with_fallback
        else:
            return llm | output_parser

    @property
    def _as_converse(self) -> ChatBedrockConverse:
        kwargs = {
            k: v
            for k, v in (self.model_kwargs or {}).items()
            if k
            in (
                "stop",
                "stop_sequences",
                "max_tokens",
                "temperature",
                "top_p",
                "additional_model_request_fields",
                "additional_model_response_field_paths",
                "performance_config",
                "request_metadata",
                "service_tier",
            )
        }
        if self.max_tokens:
            kwargs["max_tokens"] = self.max_tokens
        if self.temperature is not None:
            kwargs["temperature"] = self.temperature
        if self.stop_sequences:
            kwargs["stop_sequences"] = self.stop_sequences
        if self.service_tier:
            kwargs["service_tier"] = self.service_tier

        return ChatBedrockConverse(
            client=self.client,
            model=self.model_id,
            region_name=self.region_name,
            credentials_profile_name=self.credentials_profile_name,
            aws_access_key_id=self.aws_access_key_id,
            aws_secret_access_key=self.aws_secret_access_key,
            aws_session_token=self.aws_session_token,
            config=self.config,
            provider=self.provider or "",
            base_url=self.endpoint_url,
            guardrail_config=(self.guardrails if self._guardrails_enabled else None),  # type: ignore[call-arg]
            **kwargs,
        )
